from enum import IntEnum


class DataAccessResult(IntEnum):
    SUCCESS = 0
    HARDWARE_FAULT = 1
    TEMPORARY_FAILURE = 2
    READ_WRITE_DENIED = 3
    OBJECT_UNDEFINED = 4
    OBJECT_CLASS_INCONSISTENT = 9
    OBJECT_UNAVAILABLE = 11
    TYPE_UNMATCHED = 12
    SCOPE_OF_ACCESS_VIOLATED = 13
    DATA_BLOCK_UNAVAILABLE = 14
    LONG_GET_ABORTED = 15
    NO_LONG_GET_IN_PROGRESS = 16
    LONG_SET_ABORTED = 17
    NO_LONG_SET_IN_PROGRESS = 18
    DATA_BLOCK_NUMBER_INVALID = 19
    OTHER_REASON = 250


class GetResponse(IntEnum):
    NORMAL = 1
    WITH_DATABLOCK = 2
    WITH_LIST = 3


class SetResponse(IntEnum):
    NORMAL = 1
    DATABLOCK = 2
    LAST_DATABLOCK = 3
    LAST_DATABLOCK_WITH_LIST = 4
    WITH_LIST = 5


class ActionResult(IntEnum):
    SUCCESS = 0
    HARDWARE_FAULT = 1
    TEMPORARY_FAILURE = 2
    READ_WRITE_DENIED = 3
    OBJECT_UNDEFINED = 4
    OBJECT_CLASS_INCONSISTENT = 9
    OBJECT_UNAVAILABLE = 11
    TYPE_UNMATCHED = 12
    SCOPE_OF_ACCESS_VIOLATED = 13
    DATA_BLOCK_UNAVAILABLE = 14
    LONG_ACTION_ABORTED = 15
    NO_LONG_ACTION_IN_PROGRESS = 16
    OTHER_REASON = 250


class ActionResponse(IntEnum):
    NORMAL = 1
    WITH_PBLOCK = 2
    WITH_LIST = 3
    NEXT_PBLOCK = 4



UNKNOWN = "Неизвестное имя класса"
FOR = "для"
BILLING_PERIOD = "За расчетный период"
OVER_VOLTAGE_COUNTER = "Количество перенапряжений"

# DLMS UA 1000-1 Ed 14 7.3.2 Value group B Table 56 
INSTANCE = "экземпляр"
CHANNEL = "канал"
UTILITY_SPECIFIC = "код утилиты"
MANUFACTURER_SPECIFIC = "код производителя"
RESERVED = "зарезервирован"

GLOBAL_METER_RESET_SCRIPT_TABLE = "Таблица сценариев сброса счетчика"
MDI_RESET_END_OF_BILLING_PERIOD_SCRIPT_TABLE = "Таблица сценариев. Конец расчетного периода"
TARIFFICATION_SCRIPT_TABLE = "Таблица сценариев. Тарификация"
SET_OUTPUT_SIGNALS_SCRIPT_TABLE = "Таблица сценариев. Управление выходами"
DISCONNECT_CONTROL_SCRIPT_TABLE = "Таблица сценариев управления отключением"
IMAGE_ACTIVATION_SCRIPT_TABLE = "Таблица сценариев. Активирование перепрошивки"
PUSH_SCRIPT_TABLE = "Таблица сценарие. Инициативный выход"
RU_STOP_FRAME_SCRIPT_TABLE = "Таблица сценариев. Стоп-кадр"
SPECIAL_DAYS_TABLE = "Таблица особых дней"
END_OF_BILLING_PERIOD_SINGLE_ACTION_SCHEDULE = "Расчетный день и час"
DISCONNECT_CONTROL_SINGLE_ACTION_SCHEDULE = "Управление отключением"
IMAGE_ACTIVATION_SINGLE_ACTION_SCHEDULE = "Активирование перепрошивки"
OUTPUT_CONTROL_SINGLE_ACTION_SCHEDULE = "Управление выходами"
PUSH_SINGLE_ACTION_SCHEDULE = "Инициативный выход"
LOAD_PROFILE_CONTROL_SINGLE_ACTION_SCHEDULE = "Load profile control Single action schedule"  # TODO: add RUSSIAN
M_BUS_PROFILE_CONTROL_SINGLE_ACTION_SCHEDULE = "M-Bus profile control Single action schedule"  # TODO: add RUSSIAN
FUNCTION_CONTROL_SINGLE_ACTION_SCHEDULE = "Function control Single action schedule"  # TODO: add RUSSIAN
RU_LIMITER_BY_POWER = "Ограничитель по мощности"
RU_LIMITER_BY_CURRENT = "Ограничитель по току"
RU_LIMITER_BY_VOLTAGE = "Ограничитель по напряжению"
RU_LIMITER_BY_MAGNETIC = "Ограничитель по воздействию магнитного поля"
RU_LIMITER_BY_DIFFERENCE_CURRENT = "Ограничитель по дифференциальному току"
RU_LIMITER_BY_TEMPERATURE = "Ограничитель по превышению температуры"
RU_IEC_HDLC_SETUP_OPTO = "Настройки HDLC Оптопорт"
RU_IEC_HDLC_SETUP_RS_485 = "Настройки HDLC RS-485"
RU_IEC_HDLC_SETUP_GSM = "Настройки HDLC GSM"
CURRENT_ASSOCIATION = "Текущее соединение"
RU_PUBLIC_CLIENT_ASSOCIATION = "Соединение типа <Публичный клиент>"
RU_METER_READER_ASSOCIATION = "Соединение типа <Считыватель показаний>"
RU_UTILITY_SETTING_ASSOCIATION = "Соединение типа <Конфигуратор>"
RU_PUSH = "Соединение типа <Инициативный>"
COSEM_LOGICAL_DEVICE_NAME = "Логическое имя устройства"
INVOCATION_COUNTER = "Счетчик вызовов"
RU_SPECIFIC_PASSPORT_DATA_PROFILE = "Профиль паспортных данных"
RU_DEVICE_FACTORY_NUMBER = "Серийный номер ПУ"
RU_DEVICE_TYPE = "Тип ПУ"
RU_DEVICE_METROLOGICAL_VERSION = "Версия метрологического ПО"
RU_PRODUCER_NAME = "Наименование производителя"
RU_DEVICE_RELEASE_DATE = "Дата выпуска ПУ"
RU_REMOTE_CONSOLE_SERIAL_NUMBER = "Серийный номер пульта"
RU_SPODES_VERSION = "Версия спецификации СПОДЭС"
RU_DEVICE_CONNECTION_SCHEME = "Схема подключения ПУ"
RU_DEVICE_NOT_METROLOGICAL_VERSION = "Версия метрологического не значимого ПО"
RU_DEVICE_ID = "Идентификатор исполнения счетчика"
RU_COUNTER_POINT_DATA = "данные точки учета"
I_O_CONTROL_SIGNAL_OBJECTS_GLOBAL = "Статус входов/выходов"
POWER_QUALITY_STATUS_DAILY = "Статус качества сети (профиль суточных показаний)"
POWER_QUALITY_STATUS_QUALITY_LOG = "Статус качества сети (журнал качества сети)"
TIME_OF_OPERATION = "Время работы ПУ"
RU_DURATION_OF_FAILURE_OVERSTRAIN = "Длительность провала/перенапряжения"
RU_AMBIENT_TEMPERATURE = "Температура"
EVENTS_RELATED_TO_VOLTAGE = "События, связанные с напряжением"
RU_EVENTS_RELATED_TO_CURRENT = "События, связанные с током"
RU_EVENTS_RELATED_TO_LOAD_RELAY = "События, связанные с вкл./выкл. ПУ, коммутации реле нагрузки"
RU_EVENTS_FOR_PROGRAMMING_DEVICE_PARAMETERS = "События программирования параметров ПУ"
RU_EXTERNAL_IMPACT_EVENTS = "События внешних воздействий"
RU_COMMUNICATION_EVENTS = "Коммуникационные события"
RU_ACCESS_CONTROL_EVENTS = "События контроля доступа"
RU_EVENT_CODES_FOR_THE_SELF_DIAGNOSIS_LOG = "Коды событий для журнала самодиагностики"
RU_EVENTS_FOR_EXCEEDING_THE_REACTIVE_POWER = "События по превышению реактивной мощности tg(ф)"
RU_CHANNEL_NUMBER_INTERFACE = "Номер канала (интерфейс)"
RU_CHANGE_LIMIT_LEVEL = "Изменение уровня лимита"
COMMUNICATION_ADDRESS = "Адрес (клиента)"
LOAD_PARAMETERS_PROFILE = "Параметры профиля нагрузки"
# 6.2.4 Other abstract general purpose OBIS codes
# Program entries Table 60
ACTIVE_FIRMWARE_IDENTIFIER = "Активный идентификатор прошивки"
ACTIVE_FIRMWARE_VERSION = "Активная версия прошивки"
ACTIVE_FIRMWARE_SIGNATURE = "Активная подпись прошивки"
# 6.2.13 Register monitor and alarm monitor objects
RU_ALARM_MONITOR_1 = "Монитор событий реле нагрузки"
# 6.2.19 Standard readout profile objects
GENERAL_DISPLAY_READOUT = "Профиль основного дисплея «в Автопрокрутке»"
ALTERNATE_DISPLAY_READOUT = "Профиль основного дисплея «по кнопке»"
# 6.2.44 Parameter changes and calibration objects
PARAMETER_CHANGES_AND_CALIBRATION_0 = "Счетчик коррекций(конфигурирований)"
# 6.2.48 Status of internal control signals objects
RU_LCD_BACKLIGHT_MODE = "Режим подсветки ЖКИ"
# 6.2.59 Event counter objects
RU_RELAY_TRIGGERING_METER_FOR_OPENING = "Счетчик срабатываний реле на размыкание"
# 6.2.62 Meter tamper event related objects
METER_OPEN_EVENT_COUNTER = "Счетчик вскрытий корпуса"
METER_OPEN_EVENT_TIME_STAMP = "Дата последнего вскрытия корпуса"
METER_OPEN_EVENT_DURATION = "Продолжительность последнего вскрытия корпуса"
METER_OPEN_EVENT_CUMULATIVE_DURATION = "Общая продолжительность вскрытия корпуса"
TERMINAL_COVER_OPEN_EVENT_COUNTER = "Счетчик вскрытий крышки клеммников"
TERMINAL_COVER_OPEN_EVENT_TIME_STAMP = "Дата последнего вскрытия крышки клеммников"
TERMINAL_COVER_OPEN_EVENT_DURATION = "Продолжительность последнего вскрытия крышки клеммников"
TERMINAL_COVER_OPEN_EVENT_CUMULATIVE = "Общая продолжительность вскрытия крышки клеммников"
TILT_EVENT_COUNTER = "Счетчик наклона"
TILT_EVENT_TIME_STAMP = "Дата последнего наклона"
TILT_EVENT_DURATION = "Продолжительность последнего наклона"
TILT_EVENT_CUMULATIVE_DURATION = "Общая продолжительность наклона"
STRONG_DC_MAGNETIC_FIELD_EVENT_COUNTER = "Счетчик срабатываний датчика магнитного поля"
STRONG_DC_MAGNETIC_FIELD_EVENT_TIME_STAMP = "Дата последнего воздействия датчика магнитного поля"
STRONG_DC_MAGNETIC_FIELD_EVENT_DURATION = "Продолжительность последнего воздействия магнитного поля"
STRONG_DC_MAGNETIC_FIELD_EVENT_CUMULATIVE_DURATION = "Общая продолжительность воздействия магнитного поля"
SUPPLY_CONTROL_SWITCH_VALVE_TAMPER_EVENT_COUNTER = "Счетчик вмешательства в работу реле/клапана"
SUPPLY_CONTROL_SWITCH_VALVE_TAMPER_EVENT_TIME_STAMP = "Дата последнего в работу реле/клапана"
SUPPLY_CONTROL_SWITCH_VALVE_TAMPER_EVENT_DURATION = "Продолжительность последнего в работу реле/клапана"
SUPPLY_CONTROL_SWITCH_VALVE_TAMPER_EVENT_CUMULATIVE_DURATION = "Общая продолжительность в работу реле/клапана"
METROLOGY_TAMPER_EVENT_COUNTER = "Счетчик вмешательства в метрологическую часть"
METROLOGY_TAMPER_EVENT_TIME_STAMP = "Дата последного вмешательства в метрологическую часть"
METROLOGY_TAMPER_EVENT_DURATION = "Продолжительность последнего вмешательства в метрологическую часть"
METROLOGY_TAMPER_EVENT_CUMULATIVE_DURATION = "Общая продолжительность вмешательства в метрологическую часть"
COMMUNICATION_TAMPER_EVENT_COUNTER = "Счетчик вмешательства в коммуникационный интерфейс"
COMMUNICATION_TAMPER_EVENT_TIME_STAMP = "Дата последного вмешательства в коммуникационный интерфейс"
COMMUNICATION_TAMPER_EVENT_DURATION = "Продолжительность последнего вмешательства в коммуникационный интерфейс"
COMMUNICATION_TAMPER_EVENT_CUMULATIVE_DURATION = "Общая продолжительность вмешательства в коммуникационный интерфейс"
# 6.2.64 Alarm register, Alarm filter and Alarm descriptor objects
RU_ALARM_REGISTER_1 = "Текущее состояние инициативного выхода"
RU_ALARM_FILTER_1 = "Фильтр инициативного выхода"
RU_ALARM_REGISTER_2 = "RU. Регистр тревоги. Реле нагрузки"
RU_ALARM_FILTER_2 = "RU. Фильтр тревоги. Реле нагрузки"
# СТО 34.01-5.1-006-2021 Д.11 Журналы событий и захватываемые параметры
RU_VOLTAGE_LOG = "Журнал напряжений"
RU_CURRENT_LOG = "Журнал токов"
RU_COMMUTATION_LOG = "Журнал включений/выключений"
RU_DATA_CORRECTION_LOG = "Журнал коррекций данных"
RU_EXTERNAL_IMPACT_LOG = "Журнал внешних воздействий"
RU_COMMUNICATION_LOG = "Журнал соединений"
RU_ACCESS_LOG = "Журнал несанкционированного доступа"
RU_SELF_DIAGNOSTIC_LOG = "Журнал самодиагностики"
RU_REACTIVE_POWER_LOG = "Журнал тангенса нагрузки"
RU_QUALITY_POWER_LOG = "Журнал качества энергии"
RU_STATUS_I_O_LOG = "Журнал состояний входов/выходов"
RU_REACTIVE_POWER_LIMIT_LOG = "Журнал выхода тангенса за порог на часовом интервале"
RU_TIME_CORRECTION_LOG = "Журнал коррекции времени"
RU_START_YEAR_LOG = "Журнал на начало года"
RU_QUALITY_FOR_CALCULATION_PERIOD_LOG = "Качества сети на расчётный период"
RU_CONTROL_POWER_LOG = "Контроль мощности"
RU_BATTERY_LOG = "Батареи"
RU_CONTROL_OF_LOAD_RELAY_BLOCKER_LOG = "Контроль блокиратора реле нагрузки"
RU_TEMPERATURE_CONTROL_LOG = "Контроль температуры"
RU_VOLTAGE_DEVIATION_LOG = "«Отклонение напряжения"
RU_LINEAR_VOLTAGE_DEVIATION_LOG = "Отклонение линейного напряжения"
RU_ABNORMAL_NETWORK_SITUATION_LOG = "Нештатная ситуация сети"
RU_VOLTAGE_INTERRUPTION_LOG = "Журнал: прерывание напряжения"
RU_OVER_VOLTAGE_LOG = "Журнал: превышения напряжения"
#
ITE_FIRMWARE_DESCRIPTOR = "ИТЭ. Описание прошивки"
ITE_MAGNETIC_SENSOR_STATUS = "ИТЭ. Магнитный датчик. Статус"
ITE_DISCRETE_OUTPUTS = "ИТЭ. Дискретные выходы. Статус"
ITE_SETTING_OF_RELAY_INCLUSION_PER_DAY = "ИТЭ. Уставка количества включения реле в сутки"
ITE_SETTINGS_MESSAGES = "ИТЭ. Сообщения для настройки"
ITE_CORE_REGISTERS = "ИТЭ. Дамп регистров"
ITE_BLE_ID = "ИТЭ. BLUETOOTHLE Идентификатор"
ITE_BITMAP = "ИТЭ. BITMAP загрузчик экрана"
ITE_ICCID = "ИТЭ. Уникальный серийный номер SIM-карты"
RU_BODY_OPENING_STATE = "Текущее состояние датчика вскрытия корпуса"
RU_TERMINALS_COVER_OPENING_STATE = "Текущее состояние датчика вскрытия крышки клеммников"
RU_MAGNETIC_FIELD_STATE = "Текущее состояние датчика магнитного поля"
RU_HF_FIELD_STATE = "Текущее состояние датчика ВЧ поля"
RU_ELECTRONIC_SEALS_FIXED_STATE_OF_EVENTS = "Зафиксированное состояние событий электронных пломб"
RU_PRESSING_ELECTRONIC_SEALS = "Обжатие электронных пломб"
RU_CLEAR_OF_ELECTRONIC_SEALS_FIXED_STATE = "Очистка зафиксированных событий (магнит и ВЧ поле)"
RU_FIRST_OPENING_TIME_OF_BODY = "Время первого вскрытия электронной пломбы корпуса"
RU_FIRST_OPENING_TIME_OF_TERMINALS_COVER = "Время первого вскрытия электронной пломбы крышки клеммников"

# 7.4. Abstract object
# 7.4.1 General and service entry object - Abstract
RU_RELAY_LOAD_ARBITRATOR = "Арбитр реле нагрузки"
# Communication port log parameters
ITE_USED_COMMUNICATION_INTERFACES = "ИТЭ. Используемые коммуникационные интерфейсы"
# Consumer messages
LOCAL_CONSUMER_MESSAGE = "Потребительское сообщение через локальный информационный порт потребителей"
DISPLAY_CONSUMER_MESSAGE = "Consumer message via the meter display and / or via consumer information port"
# 7.4.5 Data profile objects – Abstract
GSM_DIAGNOSTIC_PROFILE = "Профиль GSM диагностики"
# electricity
# 7.5.1 Value group C codes – Electricity
ACTIVE_POWER_PLUS = "Положительная активная мощность (QI+QIV)"
ACTIVE_POWER_MINUS = "Отрицательная активная мощность (QII+QIII)"
REACTIVE_POWER_PLUS = "Положительная реактивная мощность (QI+QII)"
REACTIVE_POWER_MINUS = "Отрицательная реактивная мощность (QIII+QIV)"
REACTIVE_POWER_QI = "Реактивная мощность QI"
REACTIVE_POWER_QII = "Реактивная мощность QII"
REACTIVE_POWER_QIII = "Реактивная мощность QIII"
REACTIVE_POWER_QIV = "Реактивная мощность QIV"
APPARENT_POWER_PLUS = "Положительная полная мощность (QI+QIV)"
APPARENT_POWER_MINUS = "Отрицательная полная мощность (QII+QIII)"
CURRENT = "Ток"
VOLTAGE = "Напряжение"
POWER_FACTOR = "Фактор мощности"
SUPPLY_FREQUENCY = "Частота сети"
ACTIVE_POWER_ABS_PLUS = "Модуль активной мощности для измерения гармоник (abs(QI+QIV)+abs(QII+QIII))"
ACTIVE_POWER_ABS_MINUS = "Сетевая мощность (abs(QI+QIV)-abs(QII+QIII))"
ACTIVE_POWER_QI = "Активная мощность QI"
ACTIVE_POWER_QII = "Активная мощность QII"
ACTIVE_POWER_QIII = "Активная мощность QIII"
ACTIVE_POWER_QIV = "Активная мощность QIV"
REACTIVE_FACTOR = "Реактивная мощность"
CUMULATIVE = ""
ANY_PHASE = "Межфазное"
ALL_PHASE = "Все фазы"
L1 = "Фаза A"
L2 = "Фаза B"
L3 = "Фаза C"
L1_L2 = "AB"
L2_L3 = "BC"
L3_L1 = "CA"
CUMULATIVE_AMPERE_SQUARED_HOURS = "Удельная энергия потерь в цепях тока"
CUMULATIVE_VOLT_SQUARED_HOURS = "Удельная энергия потерь в силовых трансформаторах"
L0_CURRENT_NEUTRAL = "Ток нулевого провода"
RU_DIFFERENTIAL_CURRENT = "Дифференциальный ток"
PERCENT = "Проценты"
RU_LINEAR_VOLTAGE = "Линейное напряжение"
RU_POSITIVE_DEVIATION = "Положительное отклонение"
RU_NEGATIVE_DEVIATION = "Отрицательное отклонение"
POWER_REACTIVE_FACTOR = "Тангенс нагрузки"
VOLTAGE_ASYMMETRY_COEFFICIENT = "Коэффициент несимметрии напряжений"
# 7.5.2 Value group D codes – Electricity
# 7.5.2.1 Processing of measurement values
BILLING_PERIOD_AVERAGE_SINCE_LAST_RESET = "Среднее значение за расчетный период (с момента последнего сброса)"
CUMULATIVE_MINIMUM_1 = "Общий минимум 1 (с начала эксплуатации)"
CUMULATIVE_MAXIMUM_1 = "Общий максимум 1 (с начала эксплуатации)"
MINIMUM_1 = "Минимум 1 (в течение расчетного периода)"
CURRENT_AVERAGE_1 = "Текущее среднее 1 (из регистров усреднения)"
LAST_AVERAGE_1 = "Последнее среднее 1 (из регистров усреднения)"
MAXIMUM_1 = "Максимум 1 (в течение расчетного периода)"
INSTANTANEOUS_VALUE = "Мгновенное значение"
TIME_INTEGRAL_1 = "Интеграл с начала эксплуатации до текущего момента"
TIME_INTEGRAL_2 = "Интеграл с начала текущего расчетного периода"
TIME_INTEGRAL_3 = "Интеграл превышения величиной установленного порога"
CUMULATIVE_MINIMUM_2 = "Общий минимум 2"
CUMULATIVE_MAXIMUM_2 = "Общий максимум 2"
MINIMUM_2 = "Минимум 2"
CURRENT_AVERAGE_2 = "Текущее среднее 2"
LAST_AVERAGE_2 = "Последнее среднее 2"
MAXIMUM_2 = "Максимум 2"
TIME_INTEGRAL_7 = "Интеграл с начала эксплуатации до конца последнего закончившегося периода записи спериодом 1"
TIME_INTEGRAL_8 = "Интеграл с начала эксплуатации до конца последнего закончившегося периода записи спериодом 2"
TIME_INTEGRAL_9 = "Интеграл с начала текущего расчетного периода до конца последнего периода записи спериодом 1"
TIME_INTEGRAL_10 = "Интеграл с начала текущего расчетного периода до конца последнего периода записи спериодом 2"
CUMULATIVE_MINIMUM_3 = "Общий минимум 3"
CUMULATIVE_MAXIMUM_3 = "Общий максимум 3"
MINIMUM_3 = "Минимум 3"
CURRENT_AVERAGE_3 = "Текущее среднее 3"
LAST_AVERAGE_3 = "Последнее среднее 3"
MAXIMUM_3 = "Максимум 3"
CURRENT_AVERAGE_5 = "Текущее среднее 5"
CURRENT_AVERAGE_6 = "Текущее среднее 6"
TIME_INTEGRAL_5 = "Интеграл от начала текущего периода записи профиля с периодом 1 до текущего момента"
TIME_INTEGRAL_6 = "Интеграл от начала текущего периода записи профиля с периодом 2 до текущего момента"
UNDER_LIMIT_THRESHOLD = "Порог нижнего предела (провала)"
UNDER_LIMIT_OCCURRENCE_COUNTER = "Счетчик провалов"
UNDER_LIMIT_DURATION = "Продолжительность провала"
UNDER_LIMIT_MAGNITUDE = "Величина провала"
OVER_LIMIT_THRESHOLD = "Порог верхнего предела (выброса)"
OVER_LIMIT_OCCURRENCE_COUNTER = "Счетчик выбросов"
OVER_LIMIT_DURATION = "Продолжительность выброса"
OVER_LIMIT_MAGNITUDE = "Величина выброса"
MISSING_THRESHOLD = "Порог пропадания"
MISSING_OCCURRENCE_COUNTER = "Счетчик пропаданий"
MISSING_DURATION = "Продолжительность пропадания"
MISSING_MAGNITUDE = "Величина пропадания"
TIME_THRESHOLD_FOR_UNDER_LIMIT = "Порог времени для фиксации провала"
TIME_THRESHOLD_FOR_OVER_LIMIT = "Порог времени для фиксации выбросов"
TIME_THRESHOLD_FOR_MISSING_MAGNITUDE = "Порог времени для фиксации пропаданий"
CONTRACTED_VALUE = "Согласованное значение"
AVERAGE_VALUE_FOR_RECORDING_INTERVAL_1 = "?"
AVERAGE_VALUE_FOR_RECORDING_INTERVAL_2 = "?"
MINIMUM_FOR_RECORDING_INTERVAL_1 = "Минимум для периода записи 1"
MINIMUM_FOR_RECORDING_INTERVAL_2 = "Минимум для периода записи 2"
MAXIMUM_FOR_RECORDING_INTERVAL_1 = "Максимум для периода записи 1"
MAXIMUM_FOR_RECORDING_INTERVAL_2 = "Максимум для периода записи 2"
TEST_AVERAGE = "Среднее за тест"
CURRENT_AVERAGE_4_FOR_HARMONICS_MEASUREMENT = "Интеграл за время теста"
TIME_INTEGRAL_4 = "?"
RU_TOTAL_DEVIATION_TIME = "Суммарное время отклонения"
# 7.5.3.2 Tariff rates
TOTAL = "Общее"
RATE = "Тариф"
MANUFACTURER_SPECIFIC_CODES = "Код производителя"

# 7.5.3.3 Harmonics
TOTAL_FUND_ALL = "Общее"
HARMONIC = "Гармоника"
THD = "Коэффициент нелинейных искажений(THD)"
TDD = "Отношение мощности высших гармоник к максимальной мощности(TDD)"
ALL_HARMONICS = "Сумма действующих значений всех высших гармоник"
ALL_HARMONICS_TO_NOMINAL_VALUE_RATIO = "Отношение суммы действующих всех высших гармоник к номинальному значению величины"

# 7.5.3.4 Phase angles
ANGLE_FROM = "угол между"
TO = "и"
U_L1 = "Ua"
U_L2 = "Ub"
U_L3 = "Uc"
ERROR = "ошибка"
I_L1 = "Ia"
I_L2 = "Ib"
I_L3 = "Ic"
I_L0 = "N"


# 7.5.5.1 General and service entry objects – Electricity
TRANSFORMER_RATIO = "Коэффициент трансформации"
COMPLETE_COMBINED_ELECTRICITY_ID = "Полный комбинированный электрический ID"


ACTIVE_ENERGY_OUTPUT_PULSE = "Постоянная ПУ для активной энергии"
REACTIVE_ENERGY_OUTPUT_PULSE = "Постоянная ПУ для реактивной энергии"
# Ratios
TRANSFORMER_RATIO_CURRENT = F"{TRANSFORMER_RATIO} по току"
TRANSFORMER_RATIO_VOLTAGE = F"{TRANSFORMER_RATIO} по напряжению"
# Nominal values
NOMINAL_VOLTAGE = "Номинальное напряжение"
NOMINAL_CURRENT = "Номинальный ток"
NOMINAL_FREQUENCY = "Номинальная частота"
MAXIMUM_CURRENT = "Максимальный ток"
REFERENCE_VOLTAGE_FOR_POWER_QUALITY_MEASUREMENT = "Согласованное напряжение электропитания"
REFERENCED_VOLTAGE_FOR_AUX_POWER_SUPPLY = "Опорное напряжение для AUX. электропитания"
# Measurement period- / recording interval- / billing period duration
RECORDING_INTERVAL_1_FOR_LOAD_PROFILE = "Период записи в профиль 1(энергия на интервале)"
RECORDING_INTERVAL_2_FOR_LOAD_PROFILE = "Период записи в профиль 2(параметры сети)"
# Coefficients
TRANSFORMER_MAGNETIC_LOSSES = "Потери в магнитопроводе"
TRANSFORMER_IRON_LOSSES = "Потери в линии"
LINE_RESISTANCE_LOSSES = "Активное сопротивление линии"
LINE_REACTANCE_LOSSES = "Реактивное сопротивление линии"
# 7.5.5.3 List objects – Electricity
RU_MONTHLY_PROFILE = "Ежемесячный профиль"
RU_DAILY_PROFILE = "Ежесуточный профиль"

# 7.5.5.4 Data profile objects – Electricity
RU_LOAD_PROFILE = "Профиль нагрузки"

# Country specific identifiers. Russian profiles
COUNTRY_SPECIFIC_IDENTIFIER = "Страновой идентификатор"
RU_PROFILE_OF_CURRENT_VALUES = "Профиль текущих значений"
RU_SCALE_PROFILE_FOR_THE_MAGAZINE_OF_MONTHLY_INDICATIONS = "Профиль масштаба для журнала ежемесячных  показаний"
RU_SCALE_PROFILE_FOR_A_JOURNAL_OF_DAILY_INDICATION = "Профиль масштаба для журнала ежесуточных показаний"
RU_SCALE_PROFILE_FOR_CURRENT_FRAMES_OF_CURRENT_VALUES = "Профиль масштаба для стоп-кадра текущих значени"
RU_SCALE_PROFILE_FOR_LOAD_PROFILES = "Профиль масштаба для профилей нагрузки"
RU_TELEMECHANICS_PROFILE_FOR_TELEVISION_MEASUREMENTS = "Профиль телеизмерений для задач телемеханики"
RU_TELEMECHANICS_PROFILE_OF_TELEVISION_SIGNALING = "Профиль телесигнализации для задач телемеханики"
# ITE
ITE_CALIBRATION_STATUS = "ИТЭ. Статус калибровки"
ITE_CALIBRATION_APPARENT_POWER = "ИТЭ. Калибровка. Полная мощность"
ITE_CALIBRATION_ACTIVE_POWER = "ИТЭ. Калибровка. Активная мощность"
ITE_CALIBRATION_REACTIVE_POWER = "ИТЭ. Калибровка. Реактивная мощность"
ITE_CALIBRATION_VOLTAGE = "ИТЭ. Калибровка. Напряжение"
ITE_CALIBRATION_CURRENT = "ИТЭ. Калибровка. Ток"
ITE_CALIBRATION_ANGLE = "ИТЭ. Калибровка. Угол"
ITE_DISPLAY_SETTING_1 = "ИТЭ. Настройка экранов №1"
ITE_DISPLAY_SETTING_2 = "ИТЭ. Настройка экранов №2. Подменю общая энергия"
ITE_CLOCK_OFFSET_SETTING = "ИТЭ. Подстройка хода часов"
ITE_FACTORY_SETTING_10 = " ИТЭ. Производственные установки №10"
ITE_FACTORY_SETTING_11 = " ИТЭ. Производственные установки №11"
ITE_FACTORY_SETTING_12 = " ИТЭ. Производственные установки №12"
ITE_FACTORY_SETTING_13 = " ИТЭ. Производственные установки №13"
ITE_FACTORY_SETTING_14 = " ИТЭ. Производственные установки №14"
ITE_FACTORY_SETTING_15 = " ИТЭ. Производственные установки №15"
ITE_FACTORY_SETTING_16 = " ИТЭ. Производственные установки №16"
ITE_FACTORY_SETTING_17 = " ИТЭ. Производственные установки №17"
ITE_FACTORY_SETTING_18 = " ИТЭ. Производственные установки №18"
ITE_FACTORY_SETTING_19 = " ИТЭ. Производственные установки №19"

# RU. Simple OBIS. СТО 34.01-5.1-006-2021.
RU_TOTAL_VOLTAGE_DEVIATION_TIME_FOR_CALCULATED_PERIOD = "Суммарное время отклонения напряжения за расчетный период"
RU_LOAD_LOCK_STATUS = "Блокиратор реле нагрузки"  # 13.5.6



""" The OBIS identification system serves as a basis for the COSEM logical names. The system of naming COSEM objects is defined in the basic
principles (see Clause 4 EN 62056-62:2007), the identification of real data items is specified in IEC 62056-61. The following clauses define the
usage of those definitions in the COSEM environment. All codes, which are not explicitly listed, but outside the manufacturer specific range are
reserved for future use."""
from __future__ import annotations
import datetime
import dataclasses
from itertools import count, chain
from collections import deque
from functools import reduce, cached_property, lru_cache
from typing import TypeAlias, Iterator, Type
import logging
from ..version import AppVersion
from ..types import common_data_types as cdt, cosem_service_types as cst, useful_types as ut
from . import cosem_interface_class as ic
from . import events as e_
from .activity_calendar import ActivityCalendar
from .arbitrator import Arbitrator
from .association_ln.ver0 import AssociationLN as AssociationLNVer0
from .association_ln.ver1 import AssociationLN as AssociationLNVer1
from .association_ln.ver2 import AssociationLN as AssociationLNVer2
from .push_setup.ver2 import PushSetup as PushSetupVer2
from .client_setup import ClientSetup
from .clock import Clock
from .data import Data
from .disconnect_control import DisconnectControl
from .extended_register import ExtendedRegister
from .gprs_modem_setup import GPRSModemSetup
from .gsm_diagnostic import GSMDiagnostic
from .iec_hdlc_setup import IECHDLCSetup
from .image_transfer import ImageTransfer
from .ipv4_setup import IPv4Setup
from .modem_configuration.ver0 import PSTNModemConfiguration
from .modem_configuration.ver1 import ModemConfigurationVer1
from .limiter import Limiter
from .profile_generic import ProfileGeneric
from .register import Register
from .register_monitor import RegisterMonitor
from .schedule import Schedule
from .security_setup.ver0 import SecuritySetup as SecuritySetupVer0
from .security_setup.ver1 import SecuritySetup as SecuritySetupVer1
from .script_table import ScriptTable
from .single_action_schedule import SingleActionSchedule
from .special_days_table import SpecialDaysTable
from .tcp_udp_setup import TCPUDPSetup
from .. import ITE_exceptions as exc
import xml.etree.ElementTree as ET
from ..relation_to_OBIS import get_name
from ..cosem_interface_classes import implementations as impl, class_id as c_id
from .. import settings


match settings.get_current_language():
    case settings.Language.ENGLISH:        from ..Values.EN import actors
    case settings.Language.RUSSIAN:        from ..Values.RU import actors

AssociationLN: TypeAlias = AssociationLNVer0 | AssociationLNVer1 | AssociationLNVer2
AssociationLN_c: tuple[Type[AssociationLN], ...] = (AssociationLNVer0, AssociationLNVer1, AssociationLNVer2)
"""container with Association types. Use in get_instance"""
ModemConfiguration: TypeAlias = PSTNModemConfiguration | ModemConfigurationVer1
ModemConfiguration_c: tuple[Type[ModemConfiguration], ...] = (PSTNModemConfiguration, ModemConfigurationVer1)
"""container with ModemConfiguration types. Use in get_instance"""
SecuritySetup: TypeAlias = SecuritySetupVer0 | SecuritySetupVer1
SecuritySetup_c: tuple[Type[SecuritySetup], ...] = (SecuritySetupVer0, SecuritySetupVer1)
"""container with SecuritySetup types. Use in get_instance"""
PushSetup: TypeAlias = PushSetupVer2

InterfaceClass: TypeAlias = Data | Register | ProfileGeneric | Clock | ScriptTable | Schedule | SpecialDaysTable | ActivityCalendar | SingleActionSchedule | AssociationLN | \
                            IECHDLCSetup | ExtendedRegister | DisconnectControl | Limiter | ModemConfiguration | PSTNModemConfiguration | ImageTransfer | GPRSModemSetup | \
                            GSMDiagnostic | ClientSetup | SecuritySetup | TCPUDPSetup | IPv4Setup | Arbitrator | RegisterMonitor | PushSetup


def get_type_from_class(c_id: int, ver: int) -> Type[InterfaceClass]:
    match c_id:
        case 1:  return Data
        case 3:  return Register
        case 4:  return ExtendedRegister
        case 7:  return ProfileGeneric
        case 8:  return Clock
        case 9:  return ScriptTable
        case 10: return Schedule
        case 11: return SpecialDaysTable
        case 15: return AssociationLN_c[ver]
        case 18: return ImageTransfer
        case 20: return ActivityCalendar
        case 21: return RegisterMonitor
        case 22: return SingleActionSchedule
        case 23: return IECHDLCSetup
        case 27: return ModemConfiguration_c[ver]
        case 41: return TCPUDPSetup
        case 42: return IPv4Setup
        case 45: return GPRSModemSetup
        case 47: return GSMDiagnostic
        case 64: return SecuritySetup_c[ver]
        case 68: return Arbitrator
        case 70: return DisconnectControl
        case 71: return Limiter
        case 32767: return ClientSetup
        case _: raise ValueError(F"unknown DLMS class with ID {c_id}")


_CUMULATIVE = (1, 2, 11, 12, 21, 22)
_MAX_MIN_VALUES = (3, 6, 13, 16, 26, 51, 52, 53, 54)
_CURRENT_AND_LAST_AVERAGE_VALUES = (0, 4, 5, 14, 15, 24, 25, 27, 28, 49, 50, 55, 56)
_INSTANTANEOUS_VALUES = (7, 39, 41, 42)
_TIME_INTEGRAL_VALUES = (8, 9, 10, 17, 18, 19, 20, 29, 30, 58)
_OCCURRENCE_COUNTER = 40
_CONTRACTED_VALUES = (46,)
_UNDER_OVER_LIMIT_THRESHOLDS = (31, 35, 43, 44)
_UNDER_OVER_LIMIT_OCCURRENCE_COUNTERS = (32, 36)
_UNDER_OVER_LIMIT_DURATIONS = (33, 37)
_UNDER_OVER_LIMIT_MAGNITUDES = (34, 38)

_NOT_PROCESSING_OF_MEASUREMENT_VALUES = (0, 93, 94, 96, 97, 98, 99)  # BlueBook DLMS UA 1000-1 Ed.14 7.5.2.1 Table 66
_RU_CHANGE_LIMIT_LEVEL = 134

logger = logging.getLogger(__name__)
logger.level = logging.INFO


@dataclasses.dataclass(frozen=True)
class ObjectRelation:
    IC: int | tuple[int, ...] | ic.COSEMInterfaceClasses
    Additional: bytes | dict | bool = None


class Collection:
    __container: deque[InterfaceClass]
    __const_objs: list[ic.COSEMInterfaceClasses]

    def __init__(self):
        self.__container = deque()
        """ all DLMS objects container with obis key """

        self.__const_objs = list()
        """ container for major(constant) DLMS objects LN. They don't deletable """

    def __iter__(self) -> Iterator[ic.COSEMInterfaceClasses]:
        return iter(self.__container)

    @classmethod
    def from_description(cls, descriptions: list[tuple[cst.LogicalName, cdt.LongUnsigned, cdt.Unsigned]] = None):
        """ get instance with objects from descriptions """
        new_instance = cls()
        if isinstance(descriptions, list):
            deque(map(lambda description: new_instance.create(description[1], description[2], description[0]), descriptions))
        return new_instance

    def create_objects_from_collection(self, container: Collection):
        """ create objects from other collection """
        for obj in container:
            self.add_if_missing(class_id=obj.CLASS_ID,
                                version=obj.VERSION,
                                logical_name=obj.logical_name)

    def from_xml(self, filename: str, use: dict[cst.LogicalName, set[int]] = None):
        """ append objects from xml file """
        tree = ET.parse(filename)
        objects = tree.getroot()
        if use is None and objects.tag != "Configure":
            raise ValueError(F"ERROR: Root tag got {objects.tag}, expected 'Configure'")
        root_version: AppVersion = AppVersion.from_str(objects.attrib.get('version', '1.0.0'))
        logger.info(F'Версия: {root_version}, file: {filename.split("/")[-1]}')
        match root_version:
            case AppVersion(1):
                attempts: iter = count(3, -1)
                """ attempts counter """
                while len(objects) != 0 and next(attempts):
                    logger.info(F'{attempts=}')
                    for obj in objects.findall('object'):
                        new_object: ic.COSEMInterfaceClasses | None = None
                        indexes: list[int] = list()
                        """ got attributes indexes for current object """
                        for attr in obj.findall('attribute'):
                            index: ET.Element = attr.find('index')
                            if index.text.isdigit():
                                indexes.append(int(index.text))
                            else:
                                raise ValueError(F'ERROR: for {new_object.logical_name if new_object is not None else ""} got index {index.text} and it is not digital')
                            if isinstance(index, ET.Element):
                                if indexes[-1] == 1:
                                    try:
                                        ln_encoding: str = attr.findtext('encoding')
                                        if ln_encoding is None:
                                            raise ValueError('logical_name not found')
                                        class_id: str = obj.findtext('class_id')
                                        if class_id is None:
                                            raise ValueError('class_id not found')
                                        version: str = obj.findtext('version')
                                        if version is None:
                                            raise ValueError('version not found')
                                        new_object = self.add_if_missing(class_id=ut.CosemClassId(class_id),
                                                                         version=cdt.Unsigned(version),
                                                                         logical_name=cst.LogicalName(bytes.fromhex(ln_encoding)))
                                        if use is not None:
                                            use[new_object.logical_name] = set()
                                    except TypeError as e:
                                        logger.error(F'Object {obj.attrib["name"]} not created : {e}')
                                    except ValueError as e:
                                        logger.error(F'Object {obj.attrib["name"]} not created : {e}')
                                else:
                                    if new_object is None:
                                        logger.info(F'ERROR. Object "{obj.attrib["name"]}" not created : Attribute with index <1> not found')
                                        break
                                    else:
                                        try:
                                            match attr.find('encoding'), attr.find('tag'):
                                                case ET.Element() as encoding, _:
                                                    record_time: ET.Element = attr.find('record_time')
                                                    if record_time is not None:
                                                        new_object.set_record_time(indexes[-1], bytes.fromhex(record_time.text))
                                                    if new_object.logical_name == cst.LogicalName('1.0.94.7.2.255') and indexes[-1] == 2:
                                                        print('stop')
                                                    new_object.set_attr(indexes[-1], bytes.fromhex(encoding.text))
                                                    if use is not None:
                                                        use[new_object.logical_name].add(indexes[-1])
                                                case None, ET.Element() as tag:
                                                    choice = new_object.get_attr_element(indexes[-1])
                                                    if new_object.get_attr(indexes[-1]) is None and isinstance(choice, ut.CHOICE):
                                                        new_object.set_attr(indexes[-1], int(tag.text))
                                                case _ as error:
                                                    raise ValueError(F'encoding and tag not found: {error}')
                                            obj.remove(attr)
                                        except exc.NoObject as e:
                                            logger.error(F"Can't fill {new_object} attr: {indexes[-1]}. Skip. {e}.")
                                            break
                                        except exc.ITEApplication as e:
                                            logger.error(F"Can't fill {new_object} attr: {indexes[-1]}. {e}")
                                        except IndexError:
                                            logger.error(F'Object "{new_object}" not has attribute with index {index.text}')
                                        except TypeError as e:
                                            logger.error(F'Object {new_object} attr:{index.text} do not write, encoding wrong : {e}')
                                        except ValueError as e:
                                            logger.error(F'Object {new_object} attr:{index.text} do not fill: {e}')
                                        except AttributeError as e:
                                            logger.error(F'Object {new_object} attr:{index.text} do not fill: {e}')
                        if len(obj.findall('attribute')) == 1:  # only logical name
                            objects.remove(obj)
                    logger.info(F'Not parsed DLMS objects: {len(objects)}')
            case AppVersion(2, 0 | 1 | 2):
                attempts: iter = count(3, -1)
                """ attempts counter """
                while len(objects) != 0 and next(attempts):
                    logger.info(F'{attempts=}')
                    for obj in objects.findall('object'):
                        ln: str = obj.attrib.get('ln', 'is absence')
                        class_id: str = obj.findtext('class_id')
                        if not class_id:
                            logger.warning(F"skip create DLMS {ln} from Xml. Class ID is absence")
                            continue
                        version: str | None = obj.findtext('version')
                        try:
                            logical_name: cst.LogicalName = cst.LogicalName(ln)
                            if not self.is_in_collection(logical_name):
                                new_object = self.add(class_id=ut.CosemClassId(class_id),
                                                      version=None if version is None else cdt.Unsigned(version),
                                                      logical_name=cst.LogicalName(ln))
                                if use is not None:
                                    use[new_object.logical_name] = set()
                            else:
                                new_object = self.__get_object(logical_name.contents)
                        except TypeError as e:
                            logger.error(F'Object {obj.attrib["name"]} not created : {e}')
                            continue
                        except ValueError as e:
                            logger.error(F'Object {obj.attrib["name"]} not created. {class_id=} {version=} {ln=}: {e}')
                            continue
                        indexes: list[int] = list()
                        """ got attributes indexes for current object """
                        for attr in obj.findall('attribute'):
                            index: str = attr.attrib.get('index')
                            if index.isdigit():
                                indexes.append(int(index))
                            else:
                                raise ValueError(F'ERROR: for {new_object.logical_name if new_object is not None else ""} got index {index} and it is not digital')
                            try:
                                match len(attr.text), new_object.get_attr_element(indexes[-1]).DATA_TYPE:
                                    case 1 | 2, ut.CHOICE():
                                        if new_object.get_attr(indexes[-1]) is None:
                                            new_object.set_attr(indexes[-1], int(attr.text))
                                        else:
                                            """not need set"""
                                    case 1 | 2, data_type if data_type.TAG[0] == int(attr.text): """ ordering by old"""
                                    case 1 | 2, data_type:                                       raise ValueError(F'Got {attr.text} attribute Tag, expected {data_type}')
                                    case _:
                                        record_time: str = attr.attrib.get('record_time')
                                        if record_time is not None:
                                            new_object.set_record_time(indexes[-1], bytes.fromhex(record_time))
                                        new_object.set_attr(indexes[-1], bytes.fromhex(attr.text))
                                        if use is not None:
                                            use[new_object.logical_name].add(indexes[-1])
                                obj.remove(attr)
                            except exc.NoObject as e:
                                logger.error(F"Can't fill {new_object} attr: {indexes[-1]}. Skip. {e}.")
                                break
                            except exc.ITEApplication as e:
                                logger.error(F"Can't fill {new_object} attr: {indexes[-1]}. {e}")
                            except IndexError:
                                logger.error(F'Object "{new_object}" not has attr: {index}')
                            except TypeError as e:
                                logger.error(F'Object {new_object} attr:{index} do not write, encoding wrong : {e}')
                            except ValueError as e:
                                logger.error(F'Object {new_object} attr:{index} do not fill: {e}')
                            except AttributeError as e:
                                logger.error(F'Object {new_object} attr:{index} do not fill: {e}')
                        if len(obj.findall('attribute')) == 0:
                            objects.remove(obj)
                    logger.info(F'Not parsed DLMS objects: {len(objects)}')
            case _ as error:
                raise exc.VersionError(error, additional='Xml')

    def add_major(self, obj: InterfaceClass):
        self.__const_objs.append(obj)

    def get_instance(self, class_id: ut.CosemClassId,
                     version: cdt.Unsigned | None,
                     ln: cst.LogicalName) -> InterfaceClass:
        """ TODO: naming"""
        if version is None:
            version = self.set_version(class_id, version)
        match int(class_id), int(version), ln:
            case c_id.PROFILE_GENERIC,                                   1, cst.LogicalName(a, b, 99, 98, e):
                return ProfileGeneric(ln)
            case c_id.DATA,                                              0, cst.LogicalName(0 | 1, b, 0, 2, e):
                return Data(ln)
            case c_id.CLOCK,                                             0, cst.LogicalName(0, b, 1, 0, e):
                return Clock(ln)
            case c_id.MODEM_CONFIGURATION,                               v, cst.LogicalName(0, b, 2, 0, e) if v < len(ModemConfiguration_c):
                return ModemConfiguration_c[v](ln)
            case c_id.SCRIPT_TABLE,                                      0, cst.LogicalName(0, b, 10, 0, e) if e in (0, 1, 125) or 100 <= e <= 111:
                return ScriptTable(ln)
            case c_id.SPECIAL_DAYS_TABLE,                                0, cst.LogicalName(0, b, 11, 0, e):
                return SpecialDaysTable(ln)
            case c_id.SCHEDULE,                                          0, cst.LogicalName(0, b, 12, 0, e):
                return Schedule(ln)
            case c_id.ACTIVITY_CALENDAR,                                 0, cst.LogicalName(0, b, 13, 0, e):
                return ActivityCalendar(ln)
            case c_id.SINGLE_ACTION_SCHEDULE,                            0, cst.LogicalName(0, b, 15, 0, e) if 0 <= e <= 7:
                return SingleActionSchedule(ln)
            case c_id.REGISTER_MONITOR,                                  0, cst.LogicalName(0, b, 16, 0, e) | cst.LogicalName(0, b, 16, 1, e) if 0 <= e <= 9:
                return RegisterMonitor(ln)
            case c_id.LIMITER,                                           0, cst.LogicalName(0, b, 17, 0, e):
                return Limiter(ln)
            case c_id.PROFILE_GENERIC,                                   1, cst.LogicalName(0, b, 21, 0, e):
                return ProfileGeneric(ln)
            case c_id.IEC_HDLC_SETUP,                                    1, cst.LogicalName(0, b, 22, 0, 0):
                return IECHDLCSetup(ln)
            case c_id.TCP_UDP_SETUP,                                     0, cst.LogicalName(0, b, 25, 0, 0):
                return TCPUDPSetup(ln)
            case c_id.IPV4_SETUP,                                        0, cst.LogicalName(0, b, 25, 1, 0):
                return IPv4Setup(ln)
            case c_id.GPRS_MODEM_SETUP,                                  0, cst.LogicalName(0, b, 25, 4, 0):
                return GPRSModemSetup(ln)
            case c_id.GSM_DIAGNOSTIC,                                    0, cst.LogicalName(0, b, 25, 6, 0):
                return GSMDiagnostic(ln)
            case c_id.PUSH_SETUP,                                        2, cst.LogicalName(0, b, 25, 9, 0):
                return PushSetup(ln)
            case c_id.ASSOCIATION_LN_CLASS,                              v, cst.LogicalName(0, 0, 40, 0, e) if v < len(AssociationLN_c):
                return AssociationLN_c[v](ln)
            case c_id.DATA,                                              0, cst.LogicalName(0, 0, 42, 0, 0) | cst.LogicalName(0, _, 43, 1, _):
                return Data(ln)
            case c_id.SECURITY_SETUP,                                    v, cst.LogicalName(0, 0, 43, 0, e) if v < len(SecuritySetup_c):
                return SecuritySetup_c[v](ln)
            case c_id.IMAGE_TRANSFER,                                    0, cst.LogicalName(0, 0, 44, 0, e):
                return ImageTransfer(ln)
            case c_id.PROFILE_GENERIC,                                   1, cst.LogicalName(0, 0, 94, 7, 1):
                return ProfileGeneric(ln)
            case c_id.DATA,                                              0, cst.LogicalName(0, b, 96, 1, e) if 0 <= e <= 10:
                return Data(ln)
            case c_id.DATA,                                              0, cst.LogicalName(0, b, 96, 2, e):
                return Data(ln)
            case c_id.DATA,                                              0, cst.LogicalName(0, b, 96, 3, e) if 0 <= e <= 4:
                return Data(ln)
            case c_id.DISCONNECT_CONTROL,                                0, cst.LogicalName(0, b, 96, 3, 10):
                return DisconnectControl(ln)
            case c_id.ARBITRATOR,                                        0, cst.LogicalName(0, b, 96, 3, 20):
                ret = Arbitrator(ln)
                ret.actors = (actors.MANUAL,
                              actors.LOCAL_1,
                              actors.LOCAL_2,
                              actors.LOCAL_3,
                              actors.LOCAL_4,
                              actors.LOCAL_5,
                              actors.LOCAL_6,
                              actors.LOCAL_7)
                return ret
            case c_id.ARBITRATOR,                                        0, cst.LogicalName(0, b, 96, 3, e) if 21 <= e <= 29:
                return Arbitrator(ln)
            case c_id.DATA,                                              0, cst.LogicalName(0, b, 96, 4, e) if 0 <= e <= 4:
                return Data(ln)
            case c_id.DATA,                                              0, cst.LogicalName(0, b, 96, 5, e) if 0 <= e <= 4:
                return Data(ln)
            case c_id.DATA | c_id.REGISTER | c_id.EXT_REGISTER as i,     0, cst.LogicalName(0, b, 96, 8, e) if 0 <= e <= 63:
                return get_type_from_class(i, 0)(ln)
            case c_id.REGISTER | c_id.EXT_REGISTER as i,                 0, cst.LogicalName(0, b, 96, 9, e) if 0 <= e <= 2:
                return get_type_from_class(i, 0)(ln)
            case c_id.EXT_REGISTER,                                      0, cst.LogicalName(0, b, 96, 9, e) if 0 <= e <= 2:
                return ExtendedRegister(ln)
            case c_id.DATA | c_id.REGISTER | c_id.EXT_REGISTER as i,     0, cst.LogicalName(0, b, 96, 11, 0):
                ret = get_type_from_class(i, 0)(ln)
                ret.events = e_.voltage_events
                return ret
            case c_id.DATA | c_id.REGISTER | c_id.EXT_REGISTER as i,     0, cst.LogicalName(0, b, 96, 11, 1):
                ret = get_type_from_class(i, 0)(ln)
                ret.events = e_.current_events
                return ret
            case c_id.DATA | c_id.REGISTER | c_id.EXT_REGISTER as i,     0, cst.LogicalName(0, b, 96, 11, 2):
                ret = get_type_from_class(i, 0)(ln)
                ret.events = e_.commutation_events
                return ret
            case c_id.DATA | c_id.REGISTER | c_id.EXT_REGISTER as i,     0, cst.LogicalName(0, b, 96, 11, 3):
                ret = get_type_from_class(i, 0)(ln)
                ret.events = e_.programming_events
                return ret
            case c_id.DATA | c_id.REGISTER | c_id.EXT_REGISTER as i,     0, cst.LogicalName(0, b, 96, 11, 4):
                ret = get_type_from_class(i, 0)(ln)
                ret.events = e_.external_impact_events
                return ret
            case c_id.DATA | c_id.REGISTER | c_id.EXT_REGISTER as i,     0, cst.LogicalName(0, b, 96, 11, 5):
                ret = get_type_from_class(i, 0)(ln)
                ret.events = e_.communication_events
                return ret
            case c_id.DATA | c_id.REGISTER | c_id.EXT_REGISTER as i,     0, cst.LogicalName(0, b, 96, 11, 6):
                ret = get_type_from_class(i, 0)(ln)
                ret.events = e_.access_events
                return ret
            case c_id.DATA | c_id.REGISTER | c_id.EXT_REGISTER as i,     0, cst.LogicalName(0, b, 96, 11, 7):
                ret = get_type_from_class(i, 0)(ln)
                ret.events = e_.self_diagnostics_events
                return ret
            case c_id.DATA | c_id.REGISTER | c_id.EXT_REGISTER as i,     0, cst.LogicalName(0, b, 96, 11, 8):
                ret = get_type_from_class(i, 0)(ln)
                ret.events = e_.reactive_power_events
                return ret
            case c_id.DATA | c_id.REGISTER | c_id.EXT_REGISTER as i,     0, cst.LogicalName(0, b, 96, 12, e) if 0 <= e <= 6:
                return get_type_from_class(i, 0)(ln)
            case c_id.DATA,                                              0, cst.LogicalName(0, b, 96, 12, 128):
                return Data(ln)
            case c_id.DATA | c_id.REGISTER | c_id.EXT_REGISTER as i,     0, cst.LogicalName(0, b, 96, 15, e) if 0 <= e <= 99:
                return get_type_from_class(i, 0)(ln)
            case c_id.DATA | c_id.REGISTER | c_id.EXT_REGISTER as i,     0, cst.LogicalName(0, b, 96, 20, e):
                return get_type_from_class(i, 0)(ln)
            case c_id.DATA,                                              0, cst.LogicalName(0, 0, 96, 51, 0):
                return impl.data.OpeningBody(ln)
            case c_id.DATA,                                              0, cst.LogicalName(0, 0, 96, 51, 5):
                return impl.data.SealStatus(ln)
            case c_id.DATA,                                              0, cst.LogicalName(0, 0, 96, 51, e) if e in (1, 3, 4, 5, 6, 7):
                return impl.data.Unsigned(ln)
            case c_id.DATA,                                              0, cst.LogicalName(0, 0, 96, 51, e) if e == 8 or e == 9:
                return impl.data.OctetStringDateTime(ln)
            case c_id.DATA,                                              0, cst.LogicalName(0, b, 97, 98, e) if 0 <= e <= 9 or 10 <= e <= 29:
                return Data(ln)
            # 7.4.5 Data profile objects – Abstract
            case c_id.PROFILE_GENERIC,                                   1, cst.LogicalName(0, b, 99, 3, 0):
                return ProfileGeneric(ln)
            case c_id.PROFILE_GENERIC,                                   1, cst.LogicalName(0, b, 99, 1 | 2 | 12 | 13 | 14 | 15 | 16 | 17 | 18, e):
                return ProfileGeneric(ln)
            # ITE manufacture specific
            case c_id.DATA,                                              0, cst.LogicalName(0, 0, 128, 100 | 101 | 102 | 103 | 150 | 151 | 152 | 170, 0):
                return Data(ln)
            case c_id.DATA,                                              0, cst.LogicalName(0, 0, 128, 160, 0):
                return impl.data.ITEBitMap(ln)
            case c_id.CLIENT_SETUP,                                      0, cst.LogicalName(0, 0, 199, 255, 255):
                return ClientSetup(ln)
            case c_id.DATA,                                              0, cst.LogicalName(1, b, 0, 0, e) if e <= 9:
                return Data(ln)
            case c_id.DATA | c_id.REGISTER | c_id.EXT_REGISTER as i,     0, cst.LogicalName(1, b, 0, 3 | 4 | 7 | 9, e):
                return get_type_from_class(i, 0)(ln)
            # Nominal values
            case c_id.REGISTER | c_id.EXT_REGISTER as i,                 0, cst.LogicalName(1, b, 0, 6, e) if e <= 5:
                return get_type_from_class(i, 0)(ln)
            # Measurement period- / recording interval- / billing period duration
            case c_id.DATA,                 0, cst.LogicalName(1, b, 0, 8, 4 | 5):
                return impl.data.Unsigned(ln)
            # Coefficients
            case c_id.REGISTER | c_id.EXT_REGISTER as i,                 0, cst.LogicalName(1, b, 0, 10, e) if e <= 3:
                return get_type_from_class(i, 0)(ln)
            case c_id.DATA,                                              0, cst.LogicalName(1, b, 0, 11, e) if 1 <= e <= 7:
                return Data(ln)
            # RU. СТО 34.01-5.1-006-2021. 11.1 Рекомендуемые коды обозначения электрических величин
            case c_id.REGISTER,                                          0, cst.LogicalName(1, 0, 131, 35, 0):
                return Register(ln)
            case c_id.REGISTER,                                          0, cst.LogicalName(1, 0, 133, 35, 0):
                return Register(ln)
            #
            case c_id.REGISTER | c_id.EXT_REGISTER as i,                 0, cst.LogicalName(1, b, c, d, e) if c not in _NOT_PROCESSING_OF_MEASUREMENT_VALUES and \
                d in chain(_CUMULATIVE, _TIME_INTEGRAL_VALUES, _CONTRACTED_VALUES, _UNDER_OVER_LIMIT_THRESHOLDS, _UNDER_OVER_LIMIT_OCCURRENCE_COUNTERS,
                           _UNDER_OVER_LIMIT_DURATIONS, _UNDER_OVER_LIMIT_MAGNITUDES):
                return get_type_from_class(i, 0)(ln)
            case c_id.REGISTER,                                          0, cst.LogicalName(1, b, c, d, e) if c not in _NOT_PROCESSING_OF_MEASUREMENT_VALUES and \
                                                                                                              d in _INSTANTANEOUS_VALUES:
                return Register(ln)
            case c_id.REGISTER,                                          0, cst.LogicalName(1, 0, c, d, 0) if c in _CUMULATIVE and \
                                                                                                              d == _RU_CHANGE_LIMIT_LEVEL:
                return Register(ln)
            case c_id.REGISTER | c_id.EXT_REGISTER | c_id.PROFILE_GENERIC as i, 0, cst.LogicalName(1, b, c, d, e) if c not in _NOT_PROCESSING_OF_MEASUREMENT_VALUES and \
                                                                                                              d in _MAX_MIN_VALUES:
                return get_type_from_class(i, 0)(ln)
            # TODO: add DemandRegister below
            case c_id.REGISTER | c_id.EXT_REGISTER as i,                 0, cst.LogicalName(1, b, c, d, e) if c not in _NOT_PROCESSING_OF_MEASUREMENT_VALUES and \
                                                                                                              d in _CURRENT_AND_LAST_AVERAGE_VALUES:
                return get_type_from_class(i, 0)(ln)
            case c_id.DATA | c_id.REGISTER as i,                         0, cst.LogicalName(1, b, c, 40, e) if c not in _NOT_PROCESSING_OF_MEASUREMENT_VALUES:
                return get_type_from_class(i, 0)(ln)
            # RU. СТО 34.01-5.1-006-2021. 11.1 Рекомендуемые коды обозначения электрических величин
            case c_id.REGISTER,                                          0, cst.LogicalName(1, 0, 147, 133, 0):
                return Register(ln)
            case c_id.REGISTER,                                          0, cst.LogicalName(1, 0, 148, 36, 0):
                return Register(ln)
            #
            case c_id.PROFILE_GENERIC,                                1, cst.LogicalName(1, b, 94, 7, 0):
                ret = ProfileGeneric(ln)
                ret.scaler_profile_key = bytes((1, 0, 94, 7, 3, 255))
                return ret
            case c_id.PROFILE_GENERIC,                                1, cst.LogicalName(1, b, 94, 7, 1 | 2 | 3 | 4):
                return ProfileGeneric(ln)                           # Todo: RU. Scaler-profile With 1 entry
            case c_id.PROFILE_GENERIC,                                1, cst.LogicalName(1, b, 94, 7, 5 | 6):
                return ProfileGeneric(ln)                           # RU. Profile
            case c_id.PROFILE_GENERIC,                                1, cst.LogicalName(1, b, 98, 1, e):
                ret = ProfileGeneric(ln)
                ret.scaler_profile_key = bytes((1, 0, 94, 7, 1, 255))
                return ret
            case c_id.PROFILE_GENERIC,                                1, cst.LogicalName(1, b, 98, 2, e):
                ret = ProfileGeneric(ln)
                ret.scaler_profile_key = bytes((1, 0, 94, 7, 2, 255))
                return ret
            case c_id.PROFILE_GENERIC,                                1, cst.LogicalName(1, b, 99, 1 | 2, e):
                ret = ProfileGeneric(ln)
                ret.scaler_profile_key = bytes((1, 0, 94, 7, 4, 255))
                return ret
            case c_id.REGISTER,                                      0, cst.LogicalName(128, 0, c, 0, 0) if c <= 19:
                return Register(ln)
            case _:
                raise exc.NoObject(F'DLMS Object: {class_id=} {version=} {ln=} not searched in relation library')

    def add_if_missing(self, class_id: ut.CosemClassId = None,
                       version: cdt.Unsigned | None = cdt.Unsigned(),
                       logical_name: cst.LogicalName = cst.LogicalName()) -> InterfaceClass | None:
        """ like as add method with check for missing """
        if not self.is_in_collection(logical_name):
            return self.create(class_id=class_id,
                               version=version,
                               logical_name=logical_name)
        else:
            return self.__get_object(logical_name.contents)

    def get(self, obis: bytes) -> InterfaceClass | None:
        """ get object, return None if it absence """
        return next(filter(lambda obj: obj.logical_name.contents == obis, self.__container), None)

    def values(self) -> tuple[InterfaceClass]:
        return tuple(self.__container)

    def __len__(self):
        return len(self.__container)

    def create(self, class_id: ut.CosemClassId,
               version: cdt.Unsigned | None = cdt.Unsigned(0),
               logical_name: cst.LogicalName = cst.LogicalName(),
               is_major: bool = False) -> InterfaceClass:
        """ append new DLMS object in collection. <is_major>=True is not erased object """
        if self.is_in_collection(logical_name):
            raise exc.ITEApplication(F'ERROR created DLMS object with {logical_name=}. Already exist')
        new_object = self.get_instance(class_id, version, logical_name)
        new_object.collection = self
        self.__container.append(new_object)
        if is_major:
            self.__const_objs.append(new_object)
            logger.info(F'Create Major {new_object}')
        else:
            logger.info(F'Create {new_object}')
        return new_object

    def add(self, class_id: ut.CosemClassId = None,
            version: cdt.Unsigned | None = cdt.Unsigned(0),
            logical_name: cst.LogicalName = cst.LogicalName()) -> InterfaceClass:
        """ Use only in template. TODO: move to template """
        new_object = self.get_instance(class_id, version, logical_name)
        new_object.collection = self
        self.__container.append(new_object)
        logger.info(F'Create {new_object}')
        return new_object

    def raise_before(self, obj: InterfaceClass, other: InterfaceClass):
        """ Insert <obj> above <other> """
        other_index: int = self.__container.index(other)
        a = self.__container.index(obj)
        if other_index < self.__container.index(obj):
            self.__container.remove(obj)
            self.__container.insert(other_index, obj)

    def try_remove(self, logical_name: cst.LogicalName, indexes: list[int] = None) -> bool:
        """ If indexes is None when remove object else:
        Use in template. Remove attributes by indexes and remove object from collection if it has only logic attribute """
        match self.get(logical_name.contents), indexes:
            case None, _:
                return False
            case ic.COSEMInterfaceClasses() as obj, None if obj not in self.__const_objs:
                self.__container.remove(obj)
                return True
            case ic.COSEMInterfaceClasses() as obj, list():
                obj.attributes.remove_by(indexes)
                if len(obj.attributes) <= 1:
                    self.__container.remove(obj)  # TODO: Rewrite with new list attr API
                return True
            case _:
                logger.warning(F'Dont remove with: {logical_name}: {indexes=}')
                return False

    @lru_cache(maxsize=ic.get_COSEM_class_amount())
    def set_version(self, class_id: ut.CosemClassId, version: cdt.Unsigned | None = None) -> cdt.Unsigned:
        """ Set DLMS Class version for all Class ID. Return Class Version according by Class ID """
        for obj in filter(lambda it: it.CLASS_ID == class_id, self.__container):
            if version is None or version == obj.VERSION:
                return obj.VERSION
            else:
                raise ValueError(F'Not match Class Version. Expected: {obj.VERSION}, got {version}')
        if version is not None:
            return version
        else:
            raise ValueError(F'Not find version for {class_id=}')

    def is_in_collection(self, value: str | bytes | cst.LogicalName) -> bool:
        match value:
            case bytes():           obis = value
            case str():             obis = cst.LogicalName(value).contents
            case cst.LogicalName(): obis = value.contents
            case _ as error:        raise TypeError(F'Unknown type {error}')
        return obis in (obj.logical_name.contents for obj in self.__container)

    def get_object(self, value: bytes | str | cst.LogicalName | cdt.Structure | ut.CosemObjectInstanceId | ut.CosemAttributeDescriptor
                   | ut.CosemAttributeDescriptorWithSelection | ut.CosemMethodDescriptor) -> InterfaceClass:
        """ return object from obis<string> or raise exception if it absence """
        match value:
            case bytes():                                                    return self.__get_object(value)
            case cst.LogicalName() | ut.CosemObjectInstanceId():             return self.__get_object(value.contents)
            case ut.CosemAttributeDescriptor() | ut.CosemMethodDescriptor(): return self.__get_object(value.instance_id.contents)
            case ut.CosemAttributeDescriptorWithSelection():                 return self.__get_object(value.cosem_attribute_descriptor.instance_id.contents)
            case str():                                                      return self.__get_object(cst.LogicalName(value).contents)
            case cdt.Structure(logical_name=value.logical_name):             return self.__get_object(value.logical_name.contents)
            case _:                                                          raise exc.NoObject(F"Can't find DLMS Object from collection with {value=}")

    def get_first(self, values: list[str | bytes | cst.LogicalName]) -> InterfaceClass:
        """ return first object from it exist in collection from value"""
        for val in values:
            if self.is_in_collection(val):
                return self.get_object(val)
            else:
                """search next"""
        else:
            raise exc.NoObject(F"not found at least one DLMS Objects from collection with {values=}")

    def get_objects_by_class_id(self, value: int | ut.CosemClassId) -> list[InterfaceClass]:
        class_id = ut.CosemClassId(value)
        return list(filter(lambda obj: obj.CLASS_ID == class_id, self.__container))

    def get_objects_descriptions(self) -> list[tuple[cst.LogicalName, cdt.LongUnsigned, cdt.Unsigned]]:
        """ return container of objects for get device clone """
        return list(map(lambda obj: (obj.logical_name, obj.CLASS_ID, obj.VERSION), self.__container))

    def clear(self):
        """ clear to default objects amount """
        for obj in self.__container.copy():
            if obj not in self.__const_objs:
                self.__container.remove(obj)
        self.__get_object.cache_clear()

    def copy_obj_attr_values_from(self, other: InterfaceClass) -> bool:
        """ copy all attributes value from other and return bool result """
        try:
            obj: InterfaceClass = self.__get_object(other.get_obis())
            for i, attr in other.get_index_with_attributes(in_init_order=True):
                if i == 1:
                    continue
                else:
                    if attr is not None:
                        obj.set_attr(i, attr.encoding)
            return True
        except exc.NoObject as e:
            return False

    def copy_objects_attr_values_from(self, other: Collection) -> bool:
        """ Copy collections values and return True if all was writen """
        if len(other) != 0:
            return bool(reduce(lambda a, b: a or b, map(self.copy_obj_attr_values_from, other.values())))
        else:
            return False

    @property
    def current_time(self) -> datetime.datetime | None:
        return self.clock.get_current_time()

    def change_association_version(self, version: cdt.Unsigned):
        """ change Association version with clear attributes """
        del self.current_association
        logger.warning(F'Attention. ALL Association attributes will to default')
        for ass in self.get_objects_by_class_id(ut.CosemClassId(15)):
            self.__container.remove(ass)
            self.__container.append(self.get_instance(class_id=ut.CosemClassId(15),
                                                      version=version,
                                                      ln=ass.logical_name))

    @lru_cache(maxsize=256)
    def __get_object(self, obis: bytes) -> InterfaceClass:
        obj: InterfaceClass = next(filter(lambda it: it.logical_name.contents == obis, self.__container), None)
        if obj is None:
            logical_name = cst.LogicalName(bytearray(obis))
            raise exc.NoObject(F"{get_name(logical_name)}:{logical_name} is absence")
        else:
            return obj

    @cached_property
    def current_association(self) -> AssociationLN:
        return self.__get_object(bytes((0, 0, 40, 0, 0, 255)))

    @cached_property
    def PUBLIC_ASSOCIATION(self) -> AssociationLN:
        return self.__get_object(bytes((0, 0, 40, 0, 1, 255)))

    @cached_property
    def client_setup(self) -> ClientSetup:
        return self.__get_object(bytes((0, 0, 199, 255, 255, 255)))

    @property
    def clock(self) -> Clock:
        return self.__get_object(bytes((0, 0, 1, 0, 0, 255)))

    @property
    def activity_calendar(self) -> ActivityCalendar:
        return self.__get_object(bytes((0, 0, 13, 0, 0, 255)))

    @property
    def special_day_table(self) -> SpecialDaysTable:
        return self.__get_object(bytes((0, 0, 11, 0, 0, 255)))

    @property
    def IEC_HDLS_setup(self) -> IECHDLCSetup:
        return self.__get_object(bytes((0, int(self.client_setup.channel_communication), 22, 0, 0, 255)))

    @cached_property
    def TCP_UDP_setup(self) -> TCPUDPSetup:
        return self.__get_object(bytes((0, 0, 25, 0, 0, 255)))

    @lru_cache(3)
    def getIPv4Setup(self, ch: int = 0) -> IPv4Setup:
        return self.__get_object(bytes((0, ch, 25, 1, 0, 255)))

    @property
    def IPv4_setup(self) -> IPv4Setup:
        return self.__get_object(bytes((0, 0, 25, 1, 0, 255)))

    @property
    def boot_image_transfer(self) -> ImageTransfer:
        return self.__get_object(bytes((0, 0, 44, 0, 128, 255)))

    @property
    def firmware_image_transfer(self) -> ImageTransfer:
        return self.__get_object(bytes((0, 0, 44, 0, 0, 255)))

    @cached_property
    def RU_EXTENDED_PASSPORT_DATA(self) -> ProfileGeneric:
        return self.__get_object(bytes((0, 0, 94, 7, 1, 255)))

    @cached_property
    def firmware_version(self) -> Data:
        return self.__get_object(bytes((0, 0, 96, 1, 2, 255)))

    @cached_property
    def device_type(self) -> Data:
        return self.__get_object(bytes((0, 0, 96, 1, 1, 255)))

    @cached_property
    def manufacturing_date(self) -> Data:
        return self.__get_object(bytes((0, 0, 96, 1, 4, 255)))

    @cached_property
    def RU_LOAD_LOCK_STATUS(self) -> Data:
        return self.__get_object(bytes((0, 0, 96, 4, 3, 255)))

    @property
    def firmwares_description(self) -> Data:
        """ Consist from boot_version, descriptor, ex.: 0005PWRM_M2M_3_F1_5ppm_Spvq. 0.0.128.100.0.255 """
        return self.__get_object(bytes((0, 0, 128, 100, 0, 255)))

    @cached_property
    def serial_number(self) -> Data:
        """ Ex.: 0101000434322 """
        return self.__get_object(bytes((0, 0, 96, 1, 0, 255)))

    def SECURITY_SETUP(self) -> SecuritySetup:
        """SecuritySetup by order of CurrentAssociation"""
        return self.__get_object(bytes(self.current_association.security_setup_reference))

    @property
    def IC(self) -> Data:
        """ invocation counter """
        return self.__get_object(bytes((0, int(self.client_setup.channel_communication), 43, 1, self.current_association.security_setup_reference.e, 255)))

    @property
    def RU_MAGNETIC_EFFECT(self) -> Data:
        """ Russian. СПОДЕСv3 E.12.3 """
        return self.__get_object(bytes((0, 0, 96, 51, 3, 255)))

    @property
    def RU_HF_FIELD_EFFECT(self) -> Data:
        """ Russian. СПОДЕСv3 E.12.4 """
        return self.__get_object(bytes((0, 0, 96, 51, 4, 255)))

    @property
    def RU_ELECTRIC_SEAL_STATUS(self) -> impl.data.SealStatus:
        """ Russian. СПОДЕС Г.2 """
        return self.__get_object(bytes((0, 0, 96, 51, 5, 255)))

    @property
    def RU_CLOSE_ELECTRIC_SEAL(self) -> Data:
        """ Russian. СПОДЕС Г.2 """
        return self.__get_object(bytes((0, 0, 96, 51, 6, 255)))

    @property
    def RU_ERASE_MAGNETIC_EVENTS(self) -> Data:
        """ Russian. СПОДЕС Г.2 """
        return self.__get_object(bytes((0, 0, 96, 51, 7, 255)))

    @property
    def RU_ALARM_REGISTER_2(self) -> Data:
        """ Russian. Alarm register relay"""
        return self.__get_object(bytes((0, 0, 97, 98, 1, 255)))

    @property
    def RU_FILTER_ALARM_2(self) -> Data:
        """ Russian. Filter of Alarm register relay"""
        return self.__get_object(bytes((0, 0, 97, 98, 11, 255)))

    @property
    def RU_DAILY_PROFILE(self) -> ProfileGeneric:
        """ Russian. Profile of daily values """
        return self.__get_object(bytes((1, 0, 98, 2, 0, 255)))

    @cached_property
    def RU_MAXIMUM_CURRENT_EXCESS_LIMIT(self) -> Register:
        """ RU. СТО 34.01-5.1-006-2021 ver3, 11.1. Maximum current excess limit before the subscriber is disconnected, % of IMAX """
        return self.__get_object(bytes((1, 0, 11, 134, 0, 255)))

    @cached_property
    def RU_MAXIMUM_VOLTAGE_EXCESS_LIMIT(self) -> Register:
        """ RU. СТО 34.01-5.1-006-2021 ver3, 11.1. Maximum voltage excess limit before the subscriber is disconnected, % of Unominal """
        return self.__get_object(bytes((1, 0, 12, 134, 0, 255)))

    @lru_cache(5)
    def getDISCONNECT_CONTROL(self, ch: int = 0) -> DisconnectControl:
        """DLMS UA 1000-1 Ed 14 6.2.46 Disconnect control objects by channel"""
        return self.__get_object(bytes((0, ch, 96, 3, 10, 255)))

    @lru_cache(5)
    def getARBITRATOR(self, ch: int = 0) -> Arbitrator:
        """DLMS UA 1000-1 Ed 14 6.2.47 Arbitrator objects objects by channel"""
        return self.__get_object(bytes((0, ch, 96, 3, 20, 255)))

    @property
    def boot_version(self) -> str:
        try:
            return self.firmwares_description.value.to_str()[:4]
        except Exception as e:
            print(e)
            return 'unknown'

    def get_script_names(self, ln: cst.LogicalName, selector: cdt.LongUnsigned) -> str:
        """return name from script by selector"""
        obj = self.__get_object(bytes(ln))
        if isinstance(obj, ScriptTable):
            for script in obj.scripts:
                script: ScriptTable.scripts
                if script.script_identifier == selector:
                    names: list[str] = list()
                    for action in script.actions:
                        action_obj = self.__get_object(bytes(action.logical_name))
                        if int(action_obj.CLASS_ID) != int(action.class_id):
                            raise ValueError(F"got {action_obj.CLASS_ID}, expected {action.class_id}")
                        match int(action.service_id):
                            case 1:  # for write
                                if isinstance(action.parameter, cdt.NullData):
                                    names.append(action_obj.get_attr_element(int(action.index)).NAME)
                                else:
                                    raise TypeError(F"not support by framework")  # TODO: make it
                            case 2:  # for execute
                                if isinstance(action.parameter, cdt.NullData):
                                    names.append(action_obj.get_meth_element(int(action.index)).NAME)
                                else:
                                    raise TypeError(F"not support by framework")  # TODO: make it
                    return ", ".join(names)
            else:
                raise ValueError(F"not find {selector} in {obj}")
        else:
            raise ValueError(F"object with {ln} is not {ScriptTable.NAME}")



SCALER = "scaler"
UNIT = "unit"
DAY_ID = "day id"
DAY_SCHEDULE = "day schedule"
START_TIME = "start time"
END_TIME = "end time"
STOP_TIME = "stop time"
SCRIPT_LOGICAL_NAME = "script logical name"
SCRIPT_SELECTOR = "script selector"
SEASON_PROFILE_NAME = "season profile name"
SEASON_START = "season start"
WEEK_NAME = "week name"
WEEK_PROFILE_NAME = "week profile name"
MONDAY = "monday"
TUESDAY = "tuesday"
WEDNESDAY = "wednesday"
THURSDAY = "thursday"
FRIDAY = "friday"
SATURDAY = "saturday"
SUNDAY = "sunday"
ITE_OPTICAL_PORT_NAME = "optical port name"
ITE_RS485_NAME = "rs485 name"
PRESET_TIME = "preset time"
VALIDITY_INTERVAL_START = "validity interval start"
VALIDITY_INTERVAL_END = "validity interval end"
PRECEDENCE = "precedence"
DELAY = "delay"
RELIABILITY = "reliability"
PEAK_THROUGHPUT = "peak throughput"
MEAN_THROUGHPUT = "mean throughput"
DEFAULT = "default"
REQUESTED = "requested"
CELL_ID = "cell ID"
SIGNAL_QUALITY = "signal quality"
LOCATION_ID = "location ID"
BER = "ber"
IMAGE_BLOCK_NUMBER = "image block number"
IMAGE_BLOCK_VALUE = "image block value"
IMAGE_TO_ACTIVATE_SIZE = "image to activate size"
IMAGE_TO_ACTIVATE_IDENTIFICATION = "image to activate identification"
IMAGE_TO_ACTIVATE_SIGNATURE = "image to activate signature"
IMAGE_IDENTIFIER = "image identifier"
IMAGE_SIZE = "image size"
ACTION_OVER_THRESHOLD = "action over threshold"
ACTION_UNDER_THRESHOLD = "action under threshold"
EMERGENCY_PROFILE_ID = "emergency profile ID"
EMERGENCY_ACTIVATION_TIME = "emergency activation time"
EMERGENCY_DURATION = "emergency duration"
CLASS_ID = "class id"
LOGICAL_NAME = "logical name"
ATTRIBUTE_INDEX = "attribute index"
DATA_INDEX = "data index"
FROM_DATE = "from date"
TO_DATE = "to date"
FROM_ENTRY = "from entry"
TO_ENTRY = "to entry"
FROM_SELECTED_VALUE = "from selected value"
TO_SELECTED_VALUE = "to selected value"
RESTRICTING_OBJECT = "restricting object"
FROM_VALUE = "from value"
TO_VALUE = "to value"
SELECTED_VALUES = "selected values"
FIRST_INDEX = "first index"
LAST_INDEX = "last index"
FIRST_INDEX_A = "first index A"
LAST_INDEX_A = "last index A"
FIRST_INDEX_B = "first index B"
LAST_INDEX_B = "last index B"
INDEX = "index"
ENABLE = "enable"
SWITCH_TIME = "switch time"
VALIDITY_WINDOW = "validity window"
EXEC_WEEKDAYS = "exec weekdays"
EXEC_SPEC_DAYS = "exec spec days"
BEGIN_DATE = "begin date"
END_DATE = "end date"
SERVICE_ID = "service ID"
PARAMETER = "parameter"
SCRIPT_IDENTIFIER = "script identifier"
ACTIONS = "actions"
TIME = "time"
DATE = "date"
SPECIAL_DAY_DATE = "special_day_date"
REQUEST = "request"
RESPONSE = "response"
DELAY_AFTER_RESPONSE = "delay after response"
ATTRIBUTE_ACCESS = "attribute access"
METHOD_ACCESS = "method access"
JOINT_ISO_CTT_ELEMENT = "joint iso ctt element"
COUNTRY_ELEMENT = "country element"
COUNTRY_NAME_ELEMENT = "country name element"
IDENTIFIED_ORGANIZATION_ELEMENT = "identified organization element"
DLMS_UA_ELEMENT = "DLMS ua element"
APPLICATION_CONTEXT_ELEMENT = "application context element"
CONTEXT_ID_ELEMENT = "context ID element"
CLIENT_SAP = "client SAP"
SERVER_SAP = "server SAP"
ATTRIBUTE_ID = "attribute ID"
ACCESS_MODE = "access mode"
ACCESS_SELECTORS = "access selectors"
AUTHENTICATION_MECHANISM_NAME_ELEMENT = "authentication mechanism name element"
MECHANISM_ID_ELEMENT = "mechanism id element"
METHOD_ID = "method ID"
VERSION = "version"
ACCESS_RIGHTS = "access rights"
CONFORMANCE = "conformance"
MAX_RECEIVE_PDU_SIZE = "max receive pdu size"
MAX_SEND_PDU_SIZE = "max send pdu size"
DLMS_VERSION_NUMBER = "DLMS version number"
QUALITY_OF_SERVICE = "quality of service"
CYPHERING_INFO = "cyphering info"
KEY_ID = "key ID"
KEY_WRAPPED = "key wrapped"
KEY_DATA = "key data"
CERTIFICATE_ENTITY = "certificate entity"
CERTIFICATE_TYPE = "certificate type"
SERIAL_NUMBER = "serial number"
ISSUER = "issuer"
SUBJECT = "subject"
SUBJECT_ALT_NAME = "subject alt name"
SYSTEM_TITLE = "system title"
CERTIFICATE_IDENTIFICATION_TYPE = "certificate identification type"
CERTIFICATION_IDENTIFICATION_OPTIONS = "certification identification options"
REQUEST_ACTOR = "request actor"
REQUEST_ACTION_LIST = "request action list"
ACTION_UP = "action up"
ACTION_DOWN = "action down"
ITE_BIT_MAP = "BitMap data"
RESTRICTION = "restriction"
RESTRICTION_TYPE = "restriction type"
RESTRICTION_VALUE = "restriction value"
COLUMN = "column"
TRANSPORT_SERVICE = "transport service"
DESTINATION = "destination"
MESSAGE = "message"
REPETITION_DELAY_MIN = "repetition delay min"
REPETITION_DELAY_EXPONENT = "repetition delay exponent"
REPETITION_DELAY_MAX = "repetition delay max"
CONFIRMATION_START_DATE = "confirmation start date"
CONFIRMATION_INTERVAL = "confirmation interval"
KEK_ID = "kek id"
KEY_PARAMETERS = "key parameters"
KEY_CIPHERED_DATA = "key ciphered data"
KEY_INFO_TYPE = "key info type"
KEY_INFO_OPTIONS = "key info options"
KEY_INFO = "key info"
TRANSACTION_ID = "transaction id"
ORIGINATOR_SYSTEM_TITLE = "originator system title"
RECIPIENT_SYSTEM_TITLE = "recipient system title"
OTHER_INFORMATION = "other information"
PROTECTION_TYPE = "protection type"
PROTECTION_OPTIONS = "protection options"



DATA = 1
REGISTER = 3
EXT_REGISTER = 4
DEMAND_REGISTER = 5
REGISTER_ACTIVATION = 6
PROFILE_GENERIC = 7
CLOCK = 8
SCRIPT_TABLE = 9
SCHEDULE = 10
SPECIAL_DAYS_TABLE = 11
ASSOCIATION_SN_CLASS = 12
ASSOCIATION_LN_CLASS = 15
IMAGE_TRANSFER = 18
IEC_LOCAL_PORT_SETUP = 19
ACTIVITY_CALENDAR = 20
REGISTER_MONITOR = 21
SINGLE_ACTION_SCHEDULE = 22
IEC_HDLC_SETUP = 23
IEC_TWISTED_PAIR__1__SETUP = 24
M_BUS_SLAVE_PORT_SETUP = 25
UTILITY_TABLES = 26
MODEM_CONFIGURATION = 27
AUTO_ANSWER = 28
AUTO_CONNECT = 29
COSEM_DATA_PROTECTION = 30
PUSH_SETUP = 40
TCP_UDP_SETUP = 41
IPV4_SETUP = 42
PRIME_NB_OFDM_PLC_MAC_ADDRESS_SETUP = 43
PPP_SETUP = 44
GPRS_MODEM_SETUP = 45
SMTP_SETUP = 46
GSM_DIAGNOSTIC = 47
IPV6_SETUP = 48
S_FSK_PHY_MAC_SET_UP = 50
S_FSK_ACTIVE_INITIATOR = 51
S_FSK_MAC_SYNCHRONIZATION_TIMEOUTS = 52
S_FSK_MAC_COUNTERS = 53
IEC_61334_4_32_LLC_SETUP = 55
S_FSK_REPORTING_SYSTEM_LIST = 56
ISO_IEC_8802_2_LLC_TYPE_1_SETUP = 57
ISO_IEC_8802_2_LLC_TYPE_2_SETUP = 58
ISO_IEC_8802_2_LLC_TYPE_3_SETUP = 59
REGISTER_TABLE = 61
COMPACT_DATA = 62
STATUS_MAPPING = 63
SECURITY_SETUP = 64
PARAMETER_MONITOR = 65
SENSOR_MANAGER_INTERFACE_CLASS = 67
ARBITRATOR = 68
DISCONNECT_CONTROL = 70
LIMITER = 71
M_BUS_CLIENT = 72
WIRELESS_MODE_Q_CHANNEL = 73
M_BUS_MASTER_PORT_SETUP = 74
DLMS_COSEM_SERVER_M_BUS_PORT_SETUP = 76
M_BUS_DIAGNOSTIC = 77
_61334_4_32_LLC_SSCS_SETUP = 80
PRIME_NB_OFDM_PLC_PHYSICAL_LAYER_COUNTERS = 81
PRIME_NB_OFDM_PLC_MAC_SETUP = 82
PRIME_NB_OFDM_PLC_MAC_FUNCTIONAL_PARAMETERS = 83
PRIME_NB_OFDM_PLC_MAC_COUNTERS = 84
PRIME_NB_OFDM_PLC_MAC_NETWORK_ADMINISTRATION_DATA = 85
PRIME_NB_OFDM_PLC_APPLICATION_IDENTIFICATION = 86
G3_PLC_MAC_LAYER_COUNTERS = 90
NTP_SETUP = 100
ZIGBEE_SAS_STARTUP = 101
ZIGBEE_SAS_JOIN = 102
ZIGBEE_SAS_APS_FRAGMENTATION = 103
ZIGBEE_NETWORK_CONTROL = 104
ZIGBEE_TUNNEL_SETUP = 105
ACCOUNT = 111
CREDIT_INTERFACE_CLASS = 112
CHARGE = 113
TOKEN_GATEWAY = 115
ARRAY_MANAGER = 123
COMMUNICATION_PORT_PROTECTION = 124
HS_PLC_ISO_IEC_12139_1_CPAS_SETUP = 141
HS_PLC_ISO_IEC_12139_1_IP_SSAS_SETUP = 142
HS_PLC_ISO_IEC_12139_1_HDLC_SSAS_SETUP = 143

CLIENT_SETUP = 32767



from . __class_init__ import *


class QoSElement(cdt.Structure):
    """ Params of element """
    values: tuple[cdt.Unsigned, cdt.Unsigned, cdt.Unsigned, cdt.Unsigned, cdt.Unsigned]
    ELEMENTS = (cdt.StructElement(cdt.se.PRECEDENCE, cdt.Unsigned),
                cdt.StructElement(cdt.se.DELAY, cdt.Unsigned),
                cdt.StructElement(cdt.se.RELIABILITY, cdt.Unsigned),
                cdt.StructElement(cdt.se.PEAK_THROUGHPUT, cdt.Unsigned),
                cdt.StructElement(cdt.se.MEAN_THROUGHPUT, cdt.Unsigned))

    @property
    def precedence(self) -> cdt.Unsigned:
        return self.values[0]

    @property
    def delay(self) -> cdt.Unsigned:
        return self.values[1]

    @property
    def reliability(self) -> cdt.Unsigned:
        return self.values[2]

    @property
    def peak_throughput(self) -> cdt.Unsigned:
        return self.values[3]

    @property
    def mean_throughput(self) -> cdt.Unsigned:
        return self.values[4]


class QualityOfService(cdt.Structure):
    """ Specifies the quality of service parameters. It is a structure of 2 elements:
            1: defines the default or minimum characteristics of the network concerned. These parameters have to be set to best effort value;
            2: defines the requested parameters. """
    values: tuple[QoSElement, QoSElement]
    ELEMENTS = (cdt.StructElement(cdt.se.DEFAULT, QoSElement),
                cdt.StructElement(cdt.se.REQUESTED, QoSElement))

    @property
    def default(self) -> QoSElement:
        return self.values[0]

    @property
    def requested(self) -> QoSElement:
        return self.values[1]


class GPRSModemSetup(ic.COSEMInterfaceClasses):
    """ This IC allow setting up GPRS modems, by handling all data necessary data for modem management. """
    NAME = cn.GPRS_MODEM_SETUP
    CLASS_ID = ut.CosemClassId(class_id.GPRS_MODEM_SETUP)
    VERSION = cdt.Unsigned(0)
    A_ELEMENTS = (ic.ICAElement(an.APN, cdt.OctetString),
                  ic.ICAElement(an.PIN_CODE, cdt.LongUnsigned),
                  ic.ICAElement(an.QUALITY_OF_SERVICE, QualityOfService))

    def characteristics_init(self):
        """nothing do it"""

    @property
    def APN(self) -> cdt.OctetString:
        return self.get_attr(2)

    @property
    def PIN_code(self) -> cdt.LongUnsigned:
        return self.get_attr(3)

    @property
    def quality_of_service(self) -> QualityOfService:
        return self.get_attr(4)



from ...types import common_data_types as cdt
from ...cosem_interface_classes import cosem_interface_class as ic


class ClassId(cdt.LongUnsigned):
    """ Class ID type """

    def validate(self):
        if ic.is_class_id_exist(int(self)):
            raise ValueError(F'Unknown DLMS class with ID {int(self)}')



from ite_dlms.type__ import cosem_service_types as cst
from ite_dlms.cosem_interface_classes import cosem_interface_class as ic, collection
import ITE_exceptions as exc


def get_attr_index(obj: ic.COSEMInterfaceClasses) -> list[int]:
    """ return attribute indexes for reading keep configuration """
    match obj.CLASS_ID, obj.logical_name:
        case collection.Data.CLASS_ID, _:                                                        return [2]
        case collection.Register.CLASS_ID, _:                                                    return [2, 3]
        case collection.ExtendedRegister.CLASS_ID, _:                                            return [3]
        case collection.ProfileGeneric.CLASS_ID, cst.LogicalName(1, _, 94, 7, e) if 1 <= e <= 4: return [6, 3, 2, 4, 5, 8]
        case collection.ProfileGeneric.CLASS_ID, _:                                              return [6, 3, 4, 5, 8]
        case collection.Clock.CLASS_ID, _:                                                       return [8, 9]
        case collection.ScriptTable.CLASS_ID, _:                                                 return [2]
        case collection.Schedule.CLASS_ID, _:                                                    return [2]
        case collection.SpecialDaysTable.CLASS_ID, _:                                            return []
        case collection.ActivityCalendar.CLASS_ID, _:                                            return []
        case collection.SingleActionSchedule.CLASS_ID, _:                                        return [2, 3, 4]
        case collection.AssociationLNVer0.CLASS_ID, cst.LogicalName(0, 0, 40, 0, 0):             return []
        case collection.AssociationLNVer0.CLASS_ID, _:                                           return [4, 5, 7]
        case collection.IECHDLCSetup.CLASS_ID, _:                                                return [2, 3, 4, 5, 6, 7]
        case collection.DisconnectControl.CLASS_ID, _:                                           return [3, 4]
        case collection.Limiter.CLASS_ID, _:                                                     return [2, 3, 4, 5, 6, 7, 8, 9, 10, 11]
        case collection.PSTNModemConfiguration.CLASS_ID, _:                                      return [2]
        case collection.ImageTransfer.CLASS_ID, _:                                               return [2]
        case collection.GPRSModemSetup.CLASS_ID, _:                                              return [2]
        case collection.GSMDiagnostic.CLASS_ID, _:                                               return []
        case collection.ClientSetup.CLASS_ID, _:                                                 return []  # not need only for client
        case collection.TCPUDPSetup.CLASS_ID, _:                                                 return [2, 3, 4, 5, 6]
        case collection.IPv4Setup.CLASS_ID, _:                                                   return []
        case collection.Arbitrator.CLASS_ID, _:                                                  return [2]
        case collection.SecuritySetupVer0.CLASS_ID, _:                                           return [2, 3, 5]
        case collection.RegisterMonitor.CLASS_ID, _:                                             return [3, 2, 4]
        case _: raise exc.NoObject(F"Configuring. Not found {obj} with {obj.CLASS_ID} for read attributes")


def get_saved_parameters(obj: ic.COSEMInterfaceClasses) -> dict[int, int]:
    """ return attribute indexes for saved keep configuration dictionary(attr_index: 0-for value 1-for type, ...)"""
    match obj.CLASS_ID, obj.logical_name:
        case collection.Data.CLASS_ID, cst.LogicalName(0, 0, 96, 1, 1) | cst.LogicalName(0, 0, 96, 1, 3) | cst.LogicalName(0, 0, 96, 1, 6) \
                                             | cst.LogicalName(0, 0, 96, 1, 8) | cst.LogicalName(1, 0, 0, 8, 4):
            return {2: 0}
        case collection.Data.CLASS_ID, _:                                                        return {2: 1}
        case collection.Register.CLASS_ID, cst.LogicalName(1, 0, 0, 6, 0) | cst.LogicalName(1, 0, 0, 6, 1) | cst.LogicalName(1, 0, 0, 6, 2) | cst.LogicalName(1, 0, 0, 6, 3) \
                                                | cst.LogicalName(1, 0, 0, 6, 4):
            return {2: 1, 3: 1}
        case collection.Register.CLASS_ID, _:                                                    return {2: 1, 3: 0}
        case collection.ExtendedRegister.CLASS_ID, _:                                            return {2: 1, 3: 0}
        case collection.ProfileGeneric.CLASS_ID, cst.LogicalName(1, _, 94, 7, e) if 1 <= e <= 4: return {6: 0, 3: 0, 2: 0, 4: 0, 5: 0, 8: 0}
        case collection.ProfileGeneric.CLASS_ID, _:                                              return {6: 0, 3: 0, 4: 0, 5: 0, 8: 0}
        case collection.Clock.CLASS_ID, _:                                                       return {8: 0, 9: 0}
        case collection.ScriptTable.CLASS_ID, _:                                                 return {2: 0}
        case collection.Schedule.CLASS_ID, _:                                                    return {2: 0}
        case collection.SpecialDaysTable.CLASS_ID, _:                                            return dict()
        case collection.ActivityCalendar.CLASS_ID, _:                                            return dict()
        case collection.SingleActionSchedule.CLASS_ID, _:                                        return {2: 0, 3: 0, 4: 0}
        case collection.AssociationLNVer0.CLASS_ID, cst.LogicalName(0, 0, 40, 0, 0):             return {3: 0}
        case collection.AssociationLNVer0.CLASS_ID, cst.LogicalName(0, 0, 40, 0, 1):             return {2: 0, 3: 0, 4: 0, 5: 0, 6: 0, 9: 0}
        case collection.AssociationLNVer0.CLASS_ID, _:                                           return {2: 0, 3: 0, 4: 0, 5: 0, 6: 0, 7: 0, 9: 0}
        case collection.IECHDLCSetup.CLASS_ID, _:                                                return {2: 0, 3: 0, 4: 0, 5: 0, 6: 0, 7: 0, 8: 0}
        case collection.DisconnectControl.CLASS_ID, _:                                           return {3: 0, 4: 0}
        case collection.Limiter.CLASS_ID, _:                                                     return {2: 0, 3: 0, 4: 0, 5: 0, 6: 0, 7: 0, 8: 0, 9: 0, 10: 0, 11: 0}
        case collection.PSTNModemConfiguration.CLASS_ID, _:                                      return {2: 0}
        case collection.ImageTransfer.CLASS_ID, _:                                               return {2: 0}
        case collection.GPRSModemSetup.CLASS_ID, _:                                              return {2: 0}
        case collection.GSMDiagnostic.CLASS_ID, _:                                               return dict()
        case collection.ClientSetup.CLASS_ID, _:                                                 return dict()  # not need only for client
        case collection.TCPUDPSetup.CLASS_ID, _:                                                 return {2: 0, 3: 0, 4: 0, 5: 0, 6: 0}
        case collection.IPv4Setup.CLASS_ID, _:                                                   return dict()
        case collection.Arbitrator.CLASS_ID, _:                                                  return {2: 0}
        case collection.SecuritySetupVer0.CLASS_ID, _:                                           return {2: 0, 3: 0, 5: 0}
        case collection.RegisterMonitor.CLASS_ID, _:                                             return {3: 0, 2: 0, 4: 0}
        case _: raise exc.NoObject(F'Save configure. Not found {obj} with {obj.CLASS_ID} for read attributes')


if __name__ == '__main__':
    a = collection.AssociationLNVer0('0.0.1.0.0.255')
    print(a)
    b = get_attr_index(a)
    print(b)


from .__class_init__ import *
from ..types.implementations import integers


TRANSFER_NOT_INITIATED = 0
TRANSFER_INITIATED = 1
VERIFICATION_INITIATED = 2
VERIFICATION_SUCCESSFUL = 3
VERIFICATION_FAILED = 4
ACTIVATION_INITIATED = 5
ACTIVATION_SUCCESSFUL = 6
ACTIVATION_FAILED = 7


class ImageTransferStatus(cdt.Enum):
    """ Holds the status of the Image transfer process. """
    ELEMENTS = {b'\x00': en.IMAGE_TRANSFER_NOT_INITIATED,
                b'\x01': en.IMAGE_TRANSFER_INITIATED,
                b'\x02': en.IMAGE_VERIFICATION_INITIATED,
                b'\x03': en.IMAGE_VERIFICATION_SUCCESSFUL,
                b'\x04': en.IMAGE_VERIFICATION_FAILED,
                b'\x05': en.IMAGE_ACTIVATION_INITIATED,
                b'\x06': en.IMAGE_ACTIVATION_SUCCESSFUL,
                b'\x07': en.IMAGE_ACTIVATION_FAILED}


class ImageToActivateInfoElement(cdt.Structure):
    values: tuple[cdt.DoubleLongUnsigned, cdt.OctetString, cdt.OctetString]
    ELEMENTS = (cdt.StructElement(cdt.se.IMAGE_TO_ACTIVATE_SIZE, cdt.DoubleLongUnsigned),
                cdt.StructElement(cdt.se.IMAGE_TO_ACTIVATE_IDENTIFICATION, cdt.OctetString),
                cdt.StructElement(cdt.se.IMAGE_TO_ACTIVATE_SIGNATURE, cdt.OctetString))

    @property
    def image_to_activate_size(self) -> cdt.DoubleLongUnsigned:
        """size of the Image to be activated, expressed in octets"""
        return self.values[0]

    @property
    def image_to_activate_identification(self) -> cdt.OctetString:
        """identification of the Image to be activated, and may contain information like manufacturer, device type, version information, etc."""
        return self.values[1]

    @property
    def image_to_activate_signature(self) -> cdt.OctetString:
        """signature of the Image to be activated"""
        return self.values[2]


class ImageToActivateInfo(cdt.Array):
    """ Provides information on the Image(s) ready for activation. It is generated as the result of the Image verification process. Thi client may check this information before
    activating the Images. """
    TYPE = ImageToActivateInfoElement
    __getitem__: ImageToActivateInfoElement


class ImageTransferInitiate(cdt.Structure):
    """ Initializes the Image transfer process. After a successful invocation of the method the image_transfer_status attribute is set to (1) and the
    image_first_not_transferred_block_number is set to 0. Any subsequent invocation of the method resets the whole Image transfer process and all ImageBlocks need to be
    transferred again. """
    values: tuple[cdt.OctetString, cdt.DoubleLongUnsigned]
    default = (bytearray(b'default'), 0)
    ELEMENTS = (cdt.StructElement(cdt.se.IMAGE_IDENTIFIER, cdt.OctetString),
                cdt.StructElement(cdt.se.IMAGE_SIZE, cdt.DoubleLongUnsigned))

    @property
    def image_identifier(self) -> cdt.OctetString:
        """identifies the Image to be transferred. Image to be transferred (container) bur it is not necessarily linked to its content, i.e. the Images which will be activated.
        That information can be retrieved from the image_to_activate attribute after verification of the Image transferred"""
        return self.values[0]

    @property
    def image_size(self) -> cdt.DoubleLongUnsigned:
        """holds the ImageSize, expressed in octets"""
        return self.values[1]


class ImageBlockTransfer(cdt.Structure):
    """ Transfers one block of the Image to the server. After a successful invocation of the method the corresponding bit in the image_transferred_block_status attribute
    is set to 1 and the image_first_bit_transferred_block_number attribute is updated """
    values: tuple[cdt.DoubleLongUnsigned, cdt.OctetString]
    ELEMENTS = (cdt.StructElement(cdt.se.IMAGE_BLOCK_NUMBER, cdt.DoubleLongUnsigned),
                cdt.StructElement(cdt.se.IMAGE_BLOCK_VALUE, cdt.OctetString))

    @property
    def image_block_number(self) -> cdt.DoubleLongUnsigned:
        return self.values[0]

    @property
    def image_block_value(self) -> cdt.OctetString:
        return self.values[1]


class ImageTransfer(ic.COSEMInterfaceClasses):
    """ Instance of the Image transfer IC model the process of transferring binary files, called Images to COSEM servers. """
    NAME = cn.IMAGE_TRANSFER
    CLASS_ID = ut.CosemClassId(class_id.IMAGE_TRANSFER)
    VERSION = cdt.Unsigned(0)
    __blocks_for_update: list[bytearray]
    A_ELEMENTS = (ic.ICAElement(an.IMAGE_BLOCK_SIZE, cdt.DoubleLongUnsigned),
                  ic.ICAElement(an.IMAGE_TRANSFERRED_BLOCKS_STATUS, cdt.BitString),
                  ic.ICAElement(an.IMAGE_FIRST_NOT_TRANSFERRED_BLOCK_NUMBER, cdt.DoubleLongUnsigned),
                  ic.ICAElement(an.IMAGE_TRANSFER_ENABLED, cdt.Boolean),
                  ic.ICAElement(an.IMAGE_TRANSFER_STATUS, ImageTransferStatus),
                  ic.ICAElement(an.IMAGE_TO_ACTIVATE_INFO, ImageToActivateInfo))
    M_ELEMENTS = (ic.ICMElement(mn.IMAGE_TRANSFER_INITIATE, ImageTransferInitiate),
                  ic.ICMElement(mn.IMAGE_BLOCK_TRANSFER, ImageBlockTransfer),
                  ic.ICMElement(mn.IMAGE_VERIFY, integers.Only0),
                  ic.ICMElement(mn.IMAGE_ACTIVATE, integers.Only0))

    def characteristics_init(self):
        # TODO: compare image_block_size with ServerMaxReceive
        self.__blocks_for_update = list()
        """ container of blocks for transfer """

    @property
    def image_block_size(self) -> cdt.DoubleLongUnsigned:
        return self.get_attr(2)

    @property
    def image_transferred_blocks_status(self) -> cdt.BitString:
        return self.get_attr(3)

    @property
    def image_first_not_transferred_block_number(self) -> cdt.DoubleLongUnsigned:
        return self.get_attr(4)

    @property
    def image_transfer_enabled(self) -> cdt.Boolean:
        return self.get_attr(5)

    @property
    def image_transfer_status(self) -> ImageTransferStatus:
        return self.get_attr(6)

    @property
    def image_to_activate_info(self) -> ImageToActivateInfo:
        return self.get_attr(7)

    @property
    def image_transfer_initiate(self) -> ImageTransferInitiate:
        return self.get_meth(1)

    @property
    def image_block_transfer(self) -> ImageBlockTransfer:
        return self.get_meth(2)

    @property
    def image_verify(self) -> cst.Integer0:
        return self.get_meth(3)

    @property
    def image_activate(self) -> cst.Integer0:
        return self.get_meth(4)

    def set_image_for_update(self, value: bytes, identifier: bytearray):
        """ set to image_for_update from bytes and set to first block if identifier is new """
        block_size: int = self.image_block_size.decode()
        self.clear_image()
        while len(value) != 0:
            self.__blocks_for_update.append(bytearray(value[:block_size]))
            value = value[block_size:]
        if bytes(identifier) != self.image_transfer_initiate.image_identifier.contents:
            self.image_transfer_initiate.image_identifier.set(identifier)
            self.image_block_transfer.image_block_number.set(0)
        elif self.current_block_transfer == 0:
            pass
        elif self.current_block_transfer >= len(self.__blocks_for_update):
            self.image_block_transfer.image_block_number.set(0)
        else:
            self.image_block_transfer.image_block_number.set(self.current_block_transfer - 1)
        self.image_transfer_initiate.image_size.set(sum(map(len, self.__blocks_for_update)))
        self.image_block_transfer.image_block_value.set(self.__blocks_for_update[self.current_block_transfer])

    @property
    def sent_blocks_part(self) -> float:
        """ return part of sent block if available else 0 """
        if len(self.__blocks_for_update) != 0:
            return self.image_block_transfer.image_block_number.decode() / len(self.__blocks_for_update)
        else:
            return 0

    @property
    def current_block_transfer(self) -> int:
        return int(self.image_block_transfer.image_block_number)

    def set_next_block(self):
        """ set next block transfer """
        try:
            # TODO: USE bitstring of BlockStatus if it exist(length order by image) for pretty next block
            self.set_block_for_transfer(self.current_block_transfer+1)
        except IndexError:
            raise StopIteration

    def set_block_for_transfer(self, index: int):
        """ set block transfer number. IndexError if index more when blocks size """
        if index > (len(self.__blocks_for_update) - 1):
            raise IndexError(F"got block {index=}, expected 0..{len(self.__blocks_for_update)-1}")
        self.image_block_transfer.image_block_value.set(self.__blocks_for_update[index])
        self.image_block_transfer.image_block_number.set(index)

    @property
    def is_image_exist(self) -> bool:
        """ show is existing image in container """
        return bool(len(self.__blocks_for_update))

    def clear_image(self):
        """ clear image container """
        self.__blocks_for_update.clear()


if __name__ == '__main__':
    c = ImageToActivateInfo()
    c.set([[1, '00', '11']])
    a = ImageTransferInitiate(['00', 0])
    a.set(['11', 1])
    a = ImageTransferStatus()
    print(a == ImageTransferStatus('Image transfer not initiated'))
    match a:
        case ImageTransferStatus('Image transfer not initiated'): print(a)
    # print(a == 27)



from ..data import Data, ic, an, cdt, cst


class Unsigned(Data):
    """ with value type: Unsigned """
    A_ELEMENTS = ic.ICAElement(an.VALUE, cdt.Unsigned, classifier=ic.Classifier.DYNAMIC),


class OctetStringDateTime(Data):
    """ with value type: Unsigned """
    A_ELEMENTS = ic.ICAElement(an.VALUE, cst.OctetStringDateTime, classifier=ic.Classifier.DYNAMIC),


class OpeningBody(Data):
    """ RU. 0.0.96.51.0.255. СТО_34.01-5.1-006-2019v3. E 12.1 """
    A_ELEMENTS = ic.ICAElement(an.VALUE, cdt.Unsigned, classifier=ic.Classifier.DYNAMIC),


class SealUnsigned(cdt.Unsigned):
    @property
    def report(self) -> str:
        def get_name(value: int):
            """ СПОДЭСv.3 Е.12.5"""
            match value & 0b11:
                case 0: return "Не определено"
                case 1: return "Обжата"
                case 2: return "Взломана"
                case 3: return "Последущие вскрытия"
        return F"Электронные пломбы: корпуса - {get_name(int(self) & 0b11)}, крышки клеммников - {get_name((int(self) >> 2) & 0b11)}"


class SealStatus(Data):
    """ RU. 0.0.96.51.5.255. СТО_34.01-5.1-006-2019v3. E 12.1 """
    A_ELEMENTS = ic.ICAElement(an.VALUE, SealUnsigned, classifier=ic.Classifier.DYNAMIC),


class TerminalsCoverOpeningState(Data):
    """ RU. 0.0.96.51.1.255. СТО_34.01-5.1-006-2019v3. E 12.2 """
    A_ELEMENTS = ic.ICAElement(an.VALUE, cdt.Unsigned, classifier=ic.Classifier.DYNAMIC),


class BitMapData(cdt.Structure):
    values: tuple[cst.OctetStringDateTime, cst.OctetStringDateTime, cdt.OctetString]
    ELEMENTS = (cdt.StructElement(cdt.se.START_TIME, cst.OctetStringDateTime),
                cdt.StructElement(cdt.se.STOP_TIME, cst.OctetStringDateTime),
                cdt.StructElement(cdt.se.ITE_BIT_MAP, cdt.OctetString))

    @property
    def start_time(self) -> cst.OctetStringDateTime:
        """time of start show bitmap"""
        return self.values[0]

    @property
    def stop_time(self) -> cst.OctetStringDateTime:
        """time of stop show bitmap"""
        return self.values[1]

    @property
    def bitmap_data(self) -> cdt.OctetString:
        """data with bitmap format"""
        return self.values[2]


class ITEBitMap(Data):
    """ITE 0.0.128.160.0.255. Use for send struct lcd screen bitmap(BMP) with start/stop period to server"""
    A_ELEMENTS = ic.ICAElement(an.VALUE, BitMapData, classifier=ic.Classifier.STATIC),



from . import register
from .__class_init__ import *
from ..types import choices


class ExtendedRegister(register.Register, ic.COSEMInterfaceClasses):
    """ A “Register” object stores a process value or a status value with its associated unit. The register object knows
    the nature of the process value or of the status value. The nature of the value is described by the attribute
    “logical name” using the OBIS identification system. """
    NAME = cn.EXTENDED_REGISTER
    CLASS_ID = ut.CosemClassId(class_id.EXT_REGISTER)
    VERSION = cdt.Unsigned(0)
    A_ELEMENTS = (register.Register.get_attr_element(2),
                  register.Register.get_attr_element(3),
                  ic.ICAElement(an.STATUS, choices.extended_register, classifier=ic.Classifier.DYNAMIC),
                  ic.ICAElement(an.CAPTURE_TIME, cst.OctetStringDateTime, classifier=ic.Classifier.DYNAMIC))
    M_ELEMENTS = register.Register.get_meth_element(1),

    @property
    def status(self) -> cdt.CommonDataTypes:
        return self.get_attr(4)

    @property
    def capture_time(self) -> cdt.DateTime:
        return self.get_attr(5)





NULL_DATA = "null-data"
ARRAY = "array"
STRUCTURE = "structure"
BOOLEAN = "boolean"
BIT_STRING = "bit-string"
DOUBLE_LONG = "double-long"
DOUBLE_LONG_UNSIGNED = "double-long-unsigned"
OCTET_STRING = "octet-string"
VISIBLE_STRING = "visible-string"
UTF8_STRING = "utf8-string"
BCD = "bcd"
INTEGER = "integer"
LONG = "long"
UNSIGNED = "unsigned"
LONG_UNSIGNED = "long-unsigned"
COMPACT_ARRAY = "compact-array"
LONG64 = "long64"
LONG64_UNSIGNED = "unsigned-long64"
ENUM = "enum"
FLOAT32 = "float32"
FLOAT64 = "float64"
DATE_TIME = "date-time"
DATE = "date"
TIME = "time"
DON_T_CARE = "dont-care"



from ..__class_init__ import *
from ...types.implementations import structs, emuns, integers


class RestrictionValue(ut.CHOICE):
    TYPE = (cdt.Structure, cdt.NullData)
    ELEMENTS = {0: ut.SequenceElement("no restriction apply", cdt.NullData),
                1: ut.SequenceElement("restriction by date", structs.RestrictionByDate),
                2: ut.SequenceElement("restriction by entry", structs.RestrictionByEntry)}


restriction_value = RestrictionValue()


class RestrictionElement(cdt.Structure):
    """Override several methods of cdt.Structure. It limited Structure."""
    values: tuple[emuns.RestrictionType, cdt.NullData | structs.RestrictionByDate | structs.RestrictionByEntry]

    def __init__(self):
        self.__dict__["values"] = (emuns.RestrictionType(0), cdt.NullData())
        self.restriction_type.register_cb_post_set(self.__set_certificate_identification_type)

    def __set_certificate_identification_type(self):
        self.__dict__["values"] = (self.restriction_type, restriction_value(int(self.restriction_type)))

    @property
    def ELEMENTS(self) -> tuple[cdt.StructElement, cdt.StructElement]:
        match int(self.restriction_type):
            case 0: r_v = cdt.NullData
            case 1: r_v = structs.RestrictionByDate
            case 2: r_v = structs.RestrictionByEntry
            case err: raise ValueError(F"got {err} type of {self.__class__.__name__}, expect 0, 1, 2")
        return (cdt.StructElement(cdt.se.RESTRICTION_TYPE, emuns.RestrictionType),
                cdt.StructElement(cdt.se.RESTRICTION_VALUE, r_v))

    @property
    def restriction_type(self) -> emuns.RestrictionType:
        return self.values[0]

    @property
    def restriction_value(self) -> cdt.NullData | structs.RestrictionByDate | structs.RestrictionByEntry:
        return self.values[1]


class ColumnElement(cdt.Array):
    """ Specifies the list of capture objects """
    TYPE = structs.CaptureObjectDefinition


class PushObjectDefinition(cdt.Structure):
    values: tuple[cdt.LongUnsigned, cst.LogicalName, cdt.Integer, cdt.LongUnsigned, RestrictionElement, ColumnElement]
    ELEMENTS = (cdt.StructElement(cdt.se.CLASS_ID, cdt.LongUnsigned),
                cdt.StructElement(cdt.se.LOGICAL_NAME, cst.LogicalName),
                cdt.StructElement(cdt.se.ATTRIBUTE_INDEX, cdt.Integer),
                cdt.StructElement(cdt.se.DATA_INDEX, cdt.LongUnsigned),
                cdt.StructElement(cdt.se.RESTRICTION, RestrictionElement),
                cdt.StructElement(cdt.se.COLUMN, ColumnElement))

    @property
    def class_id(self) -> cdt.LongUnsigned:
        return self.values[0]

    @property
    def logical_name(self) -> cst.LogicalName:
        return self.values[1]

    @property
    def attribute_index(self) -> cdt.Integer:
        return self.values[2]

    @property
    def data_index(self) -> cdt.LongUnsigned:
        return self.values[3]

    @property
    def restriction(self) -> RestrictionElement:
        return self.values[4]

    @property
    def column(self) -> ColumnElement:
        return self.values[5]


class PushObjectList(cdt.Array):
    """ Defines the list of attributes to be pushed. """
    TYPE = PushObjectDefinition


class TransportServiceType(cdt.Enum):
    ELEMENTS = {b'\x00': en.TCP,
                b'\x01': en.UDP,
                b'\x02': en.FTP_RESERVED,
                b'\x03': en.SMTP_RESERVED,
                b'\x04': en.SMS,
                b'\x05': en.HDLC,
                b'\x06': en.M_BUS,
                b'\x07': en.ZIGBEE_RESERVED,
                b'\x08': en.DLMS_GATEWAY}
    ELEMENTS.update(((i.to_bytes(1, "big"), en.MANUFACTURER_SPECIFIC) for i in range(200, 256)))


class MessageType(cdt.Enum):
    ELEMENTS = {b'\x00': en.A_XDR,
                b'\x01': en.XML}
    ELEMENTS.update(((i.to_bytes(1, "big"), en.MANUFACTURER_SPECIFIC) for i in range(128, 256)))


class SendDestinationAndMethod(cdt.Structure):
    values: tuple[TransportServiceType, cdt.OctetString, MessageType]
    ELEMENTS = (cdt.StructElement(cdt.se.TRANSPORT_SERVICE, TransportServiceType),
                cdt.StructElement(cdt.se.DESTINATION, cdt.OctetString),
                cdt.StructElement(cdt.se.MESSAGE, MessageType))

    @property
    def transport_service(self) -> TransportServiceType:
        return self.values[0]

    @property
    def destination(self) -> cdt.OctetString:
        return self.values[1]

    @property
    def message(self) -> MessageType:
        return self.values[2]


class CommunicationWindow(cdt.Array):
    """ Defines the time points when the communication window(s) for the push become(s) active (start_time) and inactive (end_time). """
    TYPE = structs.WindowElement


class RepetitionDelay(cdt.Structure):
    values: tuple[cdt.LongUnsigned, cdt.LongUnsigned, cdt.LongUnsigned]
    ELEMENTS = (cdt.StructElement(cdt.se.REPETITION_DELAY_MIN, cdt.LongUnsigned),
                cdt.StructElement(cdt.se.REPETITION_DELAY_EXPONENT, cdt.LongUnsigned),
                cdt.StructElement(cdt.se.REPETITION_DELAY_MAX, cdt.LongUnsigned))

    @property
    def repetition_delay_min(self) -> cdt.LongUnsigned:
        return self.values[0]

    @property
    def repetition_delay_exponent(self) -> cdt.LongUnsigned:
        return self.values[1]

    @property
    def repetition_delay_max(self) -> cdt.LongUnsigned:
        return self.values[2]


class PushOperationMethod(cdt.Enum):
    ELEMENTS = {b'\x00': en.UNCONFIRMED_FAILURE,
                b'\x01': en.UNCONFIRMED_MISSING,
                b'\x02': en.CONFIRMED}


class ConfirmationParameters(cdt.Structure):
    values: tuple[cdt.DateTime, cdt.DoubleLongUnsigned]
    ELEMENTS = (cdt.StructElement(cdt.se.CONFIRMATION_START_DATE, cdt.DateTime),
                cdt.StructElement(cdt.se.CONFIRMATION_INTERVAL, cdt.DoubleLongUnsigned))

    @property
    def confirmation_start_date(self) -> cdt.DateTime:
        return self.values[0]

    @property
    def confirmation_interval(self) -> cdt.DoubleLongUnsigned:
        return self.values[1]


class IdentifiedKeyInfoOptions(cdt.Enum):
    ELEMENTS = {b'\x00': en.GLOB_UNI_ENCR_KEY,
                b'\x01': en.GLOB_BROAD_ENCR_KEY}


class KEKId(cdt.Enum):
    ELEMENTS = {b'\x00': en.MASTER_KEY}


class WrappedKeyInfoOptions(cdt.Structure):
    values: tuple[KEKId, cdt.OctetString]
    ELEMENTS = (cdt.StructElement(cdt.se.KEY_ID, KEKId),
                cdt.StructElement(cdt.se.KEY_CIPHERED_DATA, cdt.OctetString))

    @property
    def kek_id(self) -> KEKId:
        return self.values[0]

    @property
    def key_ciphered_data(self) -> cdt.OctetString:
        return self.values[1]


class AgreedKeyInfoOptions(cdt.Structure):
    values: tuple[cdt.OctetString, cdt.OctetString]
    ELEMENTS = (cdt.StructElement(cdt.se.KEY_PARAMETERS, cdt.OctetString),
                cdt.StructElement(cdt.se.KEY_CIPHERED_DATA, cdt.OctetString))

    @property
    def kek_id(self) -> cdt.OctetString:
        return self.values[0]

    @property
    def key_ciphered_data(self) -> cdt.OctetString:
        return self.values[1]


class KeyInfoOptions(ut.CHOICE):
    TYPE = (cdt.Enum, cdt.Structure)
    ELEMENTS = {0: ut.SequenceElement("identified_key", IdentifiedKeyInfoOptions),
                1: ut.SequenceElement("wrapped_key", WrappedKeyInfoOptions),
                2: ut.SequenceElement("agreed_key", AgreedKeyInfoOptions)}


key_info_options = KeyInfoOptions()


class KeyInfoElement(cdt.Structure):
    """Override several methods of cdt.Structure. It limited Structure."""
    values: tuple[emuns.KeyInfoType, IdentifiedKeyInfoOptions | WrappedKeyInfoOptions | AgreedKeyInfoOptions]

    def __init__(self):
        self.__dict__["values"] = (emuns.KeyInfoType(0), IdentifiedKeyInfoOptions())
        self.key_info_type.register_cb_post_set(self.__set_key_info_type)

    def __set_key_info_type(self):
        self.__dict__["values"] = (self.key_info_type, key_info_options(int(self.key_info_type)))

    @property
    def ELEMENTS(self) -> tuple[cdt.StructElement, cdt.StructElement]:
        match int(self.key_info_type):
            case 0: key_o = IdentifiedKeyInfoOptions
            case 1: key_o = WrappedKeyInfoOptions
            case 2: key_o = AgreedKeyInfoOptions
            case err: raise ValueError(F"got {err} type of {self.__class__.__name__}, expect 0, 1, 2")
        return (cdt.StructElement(cdt.se.KEY_INFO_TYPE, emuns.KeyInfoType),
                cdt.StructElement(cdt.se.KEY_INFO_OPTIONS, key_o))

    @property
    def key_info_type(self) -> emuns.KeyInfoType:
        return self.values[0]

    @property
    def key_info_options(self) -> IdentifiedKeyInfoOptions | WrappedKeyInfoOptions | AgreedKeyInfoOptions:
        return self.values[1]


class ProtectionOptions(cdt.Structure):
    values: tuple[cdt.OctetString, cdt.OctetString, cdt.OctetString, cdt.OctetString, KeyInfoElement]
    ELEMENTS = (cdt.StructElement(cdt.se.TRANSACTION_ID, cdt.OctetString),
                cdt.StructElement(cdt.se.ORIGINATOR_SYSTEM_TITLE, cdt.OctetString),
                cdt.StructElement(cdt.se.RECIPIENT_SYSTEM_TITLE, cdt.OctetString),
                cdt.StructElement(cdt.se.OTHER_INFORMATION, cdt.OctetString),
                cdt.StructElement(cdt.se.KEY_INFO, cdt.OctetString))

    @property
    def transaction_id(self) -> cdt.OctetString:
        return self.values[0]

    @property
    def originator_system_title(self) -> cdt.OctetString:
        return self.values[1]

    @property
    def recipient_system_title(self) -> cdt.OctetString:
        return self.values[2]

    @property
    def other_information(self) -> cdt.OctetString:
        return self.values[3]

    @property
    def key_info(self) -> KeyInfoElement:
        return self.values[4]


class ProtectionParametersElement(cdt.Structure):
    values: tuple[emuns.ProtectionType, ProtectionOptions]
    ELEMENTS = (cdt.StructElement(cdt.se.PROTECTION_TYPE, emuns.ProtectionType),
                cdt.StructElement(cdt.se.PROTECTION_OPTIONS, ProtectionOptions))

    @property
    def protection_type(self) -> emuns.ProtectionType:
        return self.values[0]

    @property
    def protection_options(self) -> ProtectionOptions:
        return self.values[1]


class PushProtectionParameters(cdt.Array):
    TYPE = ProtectionParametersElement


class PushSetup(ic.COSEMInterfaceClasses):
    """ DLMS UA 1000-1 Ed. 14 4.4.8.2 Push setup"""
    NAME = cn.PUSH_SETUP
    CLASS_ID = ut.CosemClassId(class_id.PUSH_SETUP)
    VERSION = cdt.Unsigned(2)
    A_ELEMENTS = (ic.ICAElement(an.PUSH_OBJECT_LIST, PushObjectList),
                  ic.ICAElement(an.SEND_DESTINATION_AND_METHOD, SendDestinationAndMethod),
                  ic.ICAElement(an.COMMUNICATION_WINDOW, CommunicationWindow),
                  ic.ICAElement(an.RANDOMISATION_START_INTERVAL, cdt.LongUnsigned),
                  ic.ICAElement(an.NUMBER_OF_RETRIES, cdt.Unsigned),
                  ic.ICAElement(an.REPETITION_DELAY, RepetitionDelay),
                  ic.ICAElement(an.PORT_REFERENCE, cst.LogicalName),
                  ic.ICAElement(an.PUSH_CLIENT_SAP, cdt.Integer),
                  ic.ICAElement(an.PUSH_PROTECTION_PARAMETERS, PushProtectionParameters),
                  ic.ICAElement(an.PUSH_OPERATION_METHOD, PushOperationMethod),
                  ic.ICAElement(an.CONFIRMATION_PARAMETERS, ConfirmationParameters),
                  ic.ICAElement(an.LAST_CONFIRMATION_DATE_TIME, cst.OctetStringDateTime))
    M_ELEMENTS = (ic.ICMElement(mn.PUSH, integers.Only0),
                  ic.ICMElement(mn.RESET, integers.Only0))

    def characteristics_init(self):
        """nothing do it"""

    @property
    def push_object_list(self) -> PushObjectList:
        return self.get_attr(2)

    @property
    def send_destination_and_method(self) -> SendDestinationAndMethod:
        return self.get_attr(3)

    @property
    def communication_window(self) -> CommunicationWindow:
        return self.get_attr(4)

    @property
    def randomisation_start_interval(self) -> cdt.LongUnsigned:
        return self.get_attr(5)

    @property
    def number_of_retries(self) -> cdt.Unsigned:
        return self.get_attr(6)

    @property
    def repetition_delay(self) -> RepetitionDelay:
        return self.get_attr(7)

    @property
    def port_reference(self) -> cst.LogicalName:
        return self.get_attr(8)

    @property
    def push_client_SAP(self) -> cdt.Integer:
        return self.get_attr(9)

    @property
    def push_protection_parameters(self) -> PushProtectionParameters:
        return self.get_attr(10)

    @property
    def push_operation_method(self) -> PushOperationMethod:
        return self.get_attr(11)

    @property
    def confirmation_parameters(self) -> ConfirmationParameters:
        return self.get_attr(12)

    @property
    def last_confirmation_date_time(self) -> cdt.DateTime:
        return self.get_attr(13)

    @property
    def push(self) -> integers.Only0:
        return self.get_meth(1)

    @property
    def reset(self) -> integers.Only0:
        return self.get_meth(2)



from enum import Enum, IntEnum, auto


class ITEEnum(Enum):
    def is_ok(self) -> bool:
        match self:
            case self.OK: return True
            case _:       return False

    @property
    def value(self):
        return super(ITEEnum, self).value[1]

    @property
    def importance(self):
        return super(ITEEnum, self).value[0]


class Transmit(ITEEnum):
    """ error enumerator for transmit . Priority for status has biggest value """
    OK = 0, 'Успех'
    NO_PORT = 2, 'Отсутствует порт'
    TIMEOUT = 3, 'Таймаут'
    NO_ACCESS = 4, 'Нет доступа'
    NO_TRANSPORT = 5, 'Отсутствует транспортный интерфейс'
    WRITE_ERROR = 6, 'Ошибка записи'
    READ_ERROR = 7, 'Ошибка чтения'
    EXECUTE_ERROR = 8, 'Ошибка исполнения'
    ABORT = 9, 'Прерывание'
    """ Manual interrupt """
    UNKNOWN = 10, 'Неизвестная ошибка'


class Application(ITEEnum):
    """ error enumerator for application . Priority for status has biggest value """
    OK = 0, 'Успех'
    ID_ERROR = 2, 'Ошибка серийного номера'
    """ serial number identification error """
    VERSION_ERROR = 2, 'Ошибка версии'
    """ Any errors associated with version """
    MISSING_OBJ = 2, 'Отсутствует объект'
    """ requested object was missing in collection """
    EMPTY_OBJ = 2, 'Пустые данные'
    """ empty attribute in object """
    NO_CONFIG = 2, 'Конфигурация не найдена'
    VALUE_ERROR = 2, 'Ошибка значения'
    TYPE_ERROR = 2, 'Ошибка типа прибора'
    VERIFY_ERROR = 3, 'Ошибка верификации'
    ACTIVATION_ERROR = 3, 'Ошибка активации'
    RESULT_ERROR = 3, "Ошибка данных"  # COSEMpdu_GB83.asn


class TagsName(Enum):
    ROOT = 'Configure'
    LANGUAGE = 'Language'
    GROUP = 'group'
    DEVICE = 'device'
    NAME = 'name'
    ID = 'id'
    SERIAL_NUMBER = 'serial_number'
    SERIAL_NUMBER_AMOUNT = 'serial_number_amount'
    MECHANISM_ID = 'mechanism_id'
    INTERFACE = 'Interface'
    CHILD_ID = 'child_id'
    DEVICE_ROOT = 'Objects'
    TEMPLATE_ROOT = 'template.objects'


class DBOperation(IntEnum):
    WRITE_ATTR = 1
    WRITE_METH = auto()
    CREATE_DEV_TYPE = auto()


if __name__ == '__main__':
    a = Transmit.OK
    print(a.name, a.value, a.importance)
    print(a.is_ok())
    match a:
        case Transmit.OK: print(a.value)
        case Application.ID_ERROR: print('id')



from __future__ import annotations
from .. import cosem_interface_classes
from .. import ITE_exceptions as exc
from .__class_init__ import *
from ..types import choices
from ..types.implementations import structs
threshold_scaler_unit = cdt.ScalUnitType(b'\x02\x02\x0f\x00\x16\x07')


class ValueDefinitionType(cdt.Structure):
    """ Defines an attribute of an object to be monitored. Only attributes with simple data types are allowed. """
    values: tuple[cdt.LongUnsigned, cst.LogicalName, cdt.Integer]
    ELEMENTS = (cdt.StructElement(cdt.se.CLASS_ID, cdt.LongUnsigned),
                cdt.StructElement(cdt.se.LOGICAL_NAME, cst.LogicalName),
                cdt.StructElement(cdt.se.ATTRIBUTE_INDEX, cdt.Integer))

    @property
    def class_id(self) -> cdt.LongUnsigned:
        return self.values[0]

    @property
    def logical_name(self) -> cst.LogicalName:
        return self.values[1]

    @property
    def attribute_index(self) -> cdt.Integer:
        return self.values[2]


class EmergencyProfileType(cdt.Structure):
    """ An emergency_profile is defined by three elements: emergency_profile_id, emergency_activation_time and emergency_duration.
    An emergency profile is activated if the emergency_profile_id element matches one of the elements on the emergency_profile _group_id_list, and time matches the
    emergency_activation_time and emergency_duration element """
    values: tuple[cdt.LongUnsigned, cst.OctetStringDateTime, cdt.DoubleLongUnsigned]
    ELEMENTS = (cdt.StructElement(cdt.se.EMERGENCY_PROFILE_ID, cdt.LongUnsigned),
                cdt.StructElement(cdt.se.EMERGENCY_ACTIVATION_TIME, cst.OctetStringDateTime),
                cdt.StructElement(cdt.se.EMERGENCY_DURATION, cdt.DoubleLongUnsigned))

    @property
    def emergency_profile_id(self) -> cdt.LongUnsigned:
        return self.values[0]

    @property
    def emergency_activation_time(self) -> cst.OctetStringDateTime:
        """defines the date and time when the emergency_profile activated"""
        return self.values[1]

    @property
    def emergency_duration(self) -> cdt.DoubleLongUnsigned:
        """defines the duration in seconds, for which the emergency_profile is activated"""
        return self.values[2]


class EmergencyProfileGroupIdList(cdt.Array):
    """ A list of group id-s of the emergency profile. The emergency profile can be activated only if emergency_profile_id element of the emergency_profile_type matches one of the
    elements on the emergency_profile_group_id list """
    TYPE = cdt.LongUnsigned


class ActionType(cdt.Structure):
    """ Defines the scripts to be executed when the monitored value crosses the threshold for minimal duration time. """
    values: tuple[structs.ActionItem, structs.ActionItem]
    ELEMENTS = (cdt.StructElement(cdt.se.ACTION_OVER_THRESHOLD, structs.ActionItem),
                cdt.StructElement(cdt.se.ACTION_UNDER_THRESHOLD, structs.ActionItem))

    @property
    def action_over_threshold(self) -> structs.ActionItem:
        """defines the action when the value of the attribute monitored crosses the threshold in upwards direction and remains over threshold
        for minimal over threshold duration time"""
        return self.values[0]

    def action_under_threshold(self) -> structs.ActionItem:
        """the action when the value of the attribute monitored crosses the threshold in the downwards direction and remains under threshold
        for minimal under threshold duration time."""
        return self.values[1]


class Limiter(ic.COSEMInterfaceClasses):
    """ Instances of the Limiter interface class allow defining a set of actions that are executed when the value of a value attribute of a monitored object “Data”, “Register”,
    “Extended Register”, “Demand Register”, etc. crosses the threshold value for at least minimal duration time.
        The threshold value can be normal or emergency threshold. The emergency threshold is activated via the emergency profile defined by emergency profile id, activation start
    time, and duration. The emergency profile id element is matched to an emergency profile group id: this mechanism enables the activation of the emergency threshold only
    for a specific emergency group. """
    NAME = cn.LIMITER
    CLASS_ID = ut.CosemClassId(class_id.LIMITER)
    VERSION = cdt.Unsigned(0)
    A_ELEMENTS = (ic.ICAElement(an.MONITORED_VALUE, structs.ValueDefinition),
                  ic.ICAElement(an.THRESHOLD_ACTIVE, choices.simple_dt),
                  ic.ICAElement(an.THRESHOLD_NORMAL, choices.simple_dt),
                  ic.ICAElement(an.THRESHOLD_EMERGENCY, choices.simple_dt),
                  ic.ICAElement(an.MIN_OVER_THRESHOLD_DURATION, cdt.DoubleLongUnsigned),
                  ic.ICAElement(an.MIN_UNDER_THRESHOLD_DURATION, cdt.DoubleLongUnsigned),
                  ic.ICAElement(an.EMERGENCY_PROFILE, EmergencyProfileType),
                  ic.ICAElement(an.EMERGENCY_PROFILE_GROUP_ID_LIST, EmergencyProfileGroupIdList),
                  ic.ICAElement(an.EMERGENCY_PROFILE_ACTIVE, cdt.Boolean),
                  ic.ICAElement(an.ACTIONS, ActionType))

    def characteristics_init(self):
        self._cbs_attr_post_init.update({2: lambda: self.__set_threshold_scaler_unit((3, 4, 5)),
                                         3: lambda: self.__set_threshold_scaler_unit((3,)),
                                         4: lambda: self.__set_threshold_scaler_unit((4,)),
                                         5: lambda: self.__set_threshold_scaler_unit((5,)),
                                         6: lambda: self.__set_duration_scaler_unit(6),
                                         7: lambda: self.__set_duration_scaler_unit(7)})

    @property
    def monitored_value(self) -> structs.ValueDefinition:
        return self.get_attr(2)

    @property
    def threshold_active(self) -> choices.simple_dt:
        return self.get_attr(3)

    @property
    def threshold_normal(self) -> choices.simple_dt:
        return self.get_attr(4)

    @property
    def threshold_emergency(self) -> choices.simple_dt:
        return self.get_attr(5)

    @property
    def min_over_threshold_duration(self) -> cdt.DoubleLongUnsigned:
        return self.get_attr(6)

    @property
    def min_under_threshold_duration(self) -> cdt.DoubleLongUnsigned:
        return self.get_attr(7)

    @property
    def emergency_profile(self) -> EmergencyProfileType:
        return self.get_attr(8)

    @property
    def emergency_profile_group_id_list(self) -> EmergencyProfileGroupIdList:
        return self.get_attr(9)

    @property
    def emergency_profile_active(self) -> cdt.Boolean:
        return self.get_attr(10)

    @property
    def actions(self) -> ActionType:
        return self.get_attr(11)

    def __set_duration_scaler_unit(self, attr_index: int):
        self.get_attr(attr_index).SCALER_UNIT = threshold_scaler_unit

    def __set_threshold_scaler_unit(self, attr_indexes: tuple[int, ...]):
        if self.monitored_value is not None:
            try:
                match self.collection.get_object(self.monitored_value):
                    case cosem_interface_classes.register.Register() | cosem_interface_classes.extended_register.ExtendedRegister() as obj:
                        for index in attr_indexes:
                            if self.get_attr(index) is not None:
                                self.get_attr(index).SCALER_UNIT = obj.scaler_unit
                        print('set post name')
            except exc.NoObject as e:
                print(F'For {self} threshold Scaler-unit not set. {e}')
        else:
            print('monitored_value is empty')


if __name__ == '__main__':
    a = Limiter('0.0.0.0.0.0')
    print(a)



from ..__class_init__ import *
from . import ver0


class InitializationStringElement(cdt.Structure):
    """ Request - Response strings"""
    values: tuple[cdt.OctetString, cdt.OctetString, cdt.LongUnsigned]
    ELEMENTS = (cdt.StructElement(cdt.se.REQUEST, cdt.OctetString),
                cdt.StructElement(cdt.se.RESPONSE, cdt.OctetString),
                cdt.StructElement(cdt.se.DELAY_AFTER_RESPONSE, cdt.LongUnsigned))

    @property
    def request(self) -> cdt.OctetString:
        return self.values[0]

    @property
    def response(self) -> cdt.OctetString:
        return self.values[1]

    @property
    def delay_after_response(self) -> cdt.LongUnsigned:
        return self.values[2]


class InitializationString(cdt.Array):
    """ Contains all the necessary initialization commands to be sent to the modem in order to configure it properly. This may include the configuration of special modem
    features. If the array contains more than one initialization_string_element, the requests are set in a sequence.The next request is sent after the expected response matching
    the previous request and waiting a delay-after-response time [ms], to allow the modem toe execute teh request.
    It is assumed that the modem is per-configured so that it accepts the initialization-string. If no initialization is needed, the initialization string is empty. """
    TYPE = InitializationStringElement


class ModemConfigurationVer1(ic.COSEMInterfaceClasses):
    """ This IC allow modelling the configuration and initialisation of modems used for data transfer from/to a device. Several modems can be configured."""
    NAME = cn.MODEM_CONFIGURATION
    CLASS_ID = ut.CosemClassId(27)
    VERSION = cdt.Unsigned(1)
    A_ELEMENTS = (ver0.PSTNModemConfiguration.get_attr_element(2),
                  ic.ICAElement(an.INITIALIZATION_STRING, InitializationString),
                  ver0.PSTNModemConfiguration.get_attr_element(4))

    def characteristics_init(self):
        """nothing do it"""

    @property
    def initialization_string(self) -> InitializationString:
        return self.get_attr(3)


if __name__ == '__main__':
    a = ModemConfigurationVer1('0.0.2.0.0.255')
    print(a)



from .__class_init__ import *


class IPAddress(cdt.DoubleLongUnsigned):

    def from_str(self, value: str) -> bytes:
        """ create ip: integer from string type ddd.ddd.ddd.ddd, ex.: 127.0.0.1 """
        raw_value = bytes()
        for separator in '... ':
            try:
                element, value = value.split(separator, 1)
            except ValueError:
                element, value = value, ''
            raw_value += self.__get_attr_element(element)
        return raw_value

    @staticmethod
    def __get_attr_element(value: str) -> bytes:
        if isinstance(value, str):
            if value == '':
                return b'\x00'
            try:
                return int(value).to_bytes(1, 'big')
            except OverflowError:
                raise ValueError(F'Int too big to convert {value}')
        else:
            raise TypeError(F'Unsupported type validation from string, got {value.__class__}')

    def __str__(self):
        return F'{self.contents[0]}.{self.contents[1]}.{self.contents[2]}.{self.contents[3]}'


class IPv4Setup(ic.COSEMInterfaceClasses):
    """ This IC allows modelling the setup of teh IPv4 layer, handling all information related to the IP Address settings associated to a given device adn to a lower layer
    connection on which these settings are used. There shall be and instance of this IC in a device for each different network interface implemented. For example, if a device has
    two interfaces (using the TCP-UDP/ITv4 profile on both of them), there shall be two instances of the IPc4 setup IC in that device: one for each of these interfaces."""
    NAME = cn.IPV4_SETUP
    CLASS_ID = ut.CosemClassId(class_id.IPV4_SETUP)
    VERSION = cdt.Unsigned(0)
    # TODO: more 7 attr and 3 methods
    A_ELEMENTS = (ic.ICAElement(an.DL_REFERENCE, cdt.OctetString),
                  ic.ICAElement(an.IP_ADDRESS, IPAddress),
                  # TODO: more 7 attr and 3 methods
                  )

    def characteristics_init(self):
        self.set_attr(3, 3232235521)

    @property
    def DL_reference(self) -> cdt.OctetString:
        return self.get_attr(2)

    @property
    def IP_address(self) -> IPAddress:
        return self.get_attr(3)



from ...types import common_data_types as cdt
from . import ver1


class AssociationLN(ver1.AssociationLN):
    """ COSEM logical devices able to establish application associations within a COSEM context using logical name referencing, model the associations
    through instances of the “Association LN” class. A COSEM logical device has one instance of this IC for each association
    the device is able to support"""
    VERSION = cdt.Unsigned(2)

    def characteristics_init(self):
        super(AssociationLN, self).characteristics_init()
        # TODO: more 2 attribute
        # TODO: more 2 methods


if __name__ == '__main__':
    a = AssociationLN('0.0.40.0.0.255')
    print(a.object_list)
    b = bytes.fromhex('01 01 02 04 12 00 08 11 00 09 06 00 00 01 00 00 FF 02 02 01 09 02 03 0F 01 16 01 00 02 03 0F 02 16 03 00 02 03 0F 03 16 03 00 02 03 0F 04 16 03 00 02 03 0F 05 16 03 00 02 03 0F 06 16 03 00 02 03 0F 07 16 03 00 02 03 0F 08 16 03 00 02 03 0F 09 16 03 00 01 06 02 02 0F 01 16 01 02 02 0F 02 16 01 02 02 0F 03 16 01 02 02 0F 04 16 01 02 02 0F 05 16 01 02 02 0F 06 16 01')
    a = ver1.ObjectListType(b)
    print(a)
    a.append()
    print(a)
